{ *************************************************************************** }
{                                                                             }
{ Hotzonu 2 Splitter Control Unit                                             }
{                                                                             }
{ 2001-02 www.monazilla.org Dax mailto:daxmonazilla@yahoo.co.jp               }
{                                                                             }
{ *************************************************************************** }

//  QlR|[lg
//  TCuteSplitter Version 1.10
//  Copyright (c) 1999 
//  http://www2m.biglobe.ne.jp/~fumika/   

unit ZonuSplitter;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

type
  TZonuSplitter = class(TSplitter)
  private
    { Private 錾 }
    FOpen: boolean;
    FOnOpen: TNotifyEvent;
    FOnClose: TNotifyEvent;
    FChevronRect : TRect;
    FMouseInChevron : boolean;
    FDefCursor: TCursor;
    FOldPosition: Integer;
    FChevronCursor: TCursor;
    procedure SetMouseInChevron(const Value: boolean);
    procedure ChangePane(const Open: boolean);
  protected
    { Protected 錾 }
    procedure Click; override;
    procedure Paint; override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure CMMouseLeave(var Msg: TMessage); message CM_MOUSELEAVE;
    property  MouseInChevron: boolean read  FMouseInChevron write SetMouseInChevron;
  public
    { Public 錾 }
    constructor Create(AOwner: TComponent); override;
    procedure OpenPane;
    procedure ClosePane;
  published
    { Published 錾 }
    property  ChevronCursor: TCursor  read  FChevronCursor  write FChevronCursor;
    property  OnOpen: TNotifyEvent  read  FOnOpen write FOnOpen;
    property  OnClose: TNotifyEvent read  FOnClose  write FOnClose;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Samples', [TZonuSplitter]);
end;

type
  TDummySplitter = class(TGraphicControl)
  private
    FActiveControl: TWinControl;
    FBeveled: Boolean;
    FBrush: TBrush;
    FControl: TControl;
    FDownPos: TPoint;
    FLineDC: HDC;
    FLineVisible: Boolean;
    FMinSize: NaturalNumber;
    FMaxSize: Integer;
    FNewSize: Integer;
    FOldKeyDown: TKeyEvent;
    FOldSize: Integer;
    FPrevBrush: HBrush;
    FResizeStyle: TResizeStyle;
    FSplit: Integer;
  end;

{ TZonuSplitter }

constructor TZonuSplitter.Create(AOwner: TComponent);
begin
  inherited;
  Width           :=  6;
  FMouseInChevron :=  False;
  TDummySplitter(Self).FNewSize :=  MinSize;
  FOpen           :=  True;
  FChevronCursor  :=  crDefault;
end;

procedure TZonuSplitter.Paint;
const
  CHEVRON_RECT_SIZE = 20;
  CHEVRON_SIZE = 4;
var
  BarCenter : integer;
  FOpen : boolean;
begin
  inherited Paint;
  with  Canvas  do  begin
    if  (FMouseInChevron) then  begin
      Pen.Color   :=  $00FF8080;
      Brush.Color :=  $00FF8080;
    end else begin
      Pen.Color   :=  $00804000;
      Brush.Color :=  $00804000;
    end;
    if  (Align in [alLeft, alRight])  then  begin
      BarCenter     :=  Height div 2;
      FChevronRect  :=  Rect(0,
                             BarCenter - CHEVRON_RECT_SIZE,
                             Width,
                             BarCenter + CHEVRON_RECT_SIZE);
    end else begin
      BarCenter     :=  Width div 2;
      FChevronRect  :=  Rect(BarCenter - CHEVRON_RECT_SIZE,
                             0,
                             BarCenter + CHEVRON_RECT_SIZE,
                             Height);
    end;
    FOpen :=  TDummySplitter(Self).FNewSize  >  MinSize;

    if  (Align = alLeft)  and (FOpen) or
        (Align = alRight) and (not FOpen) then
    begin
      Polygon([Point(FChevronRect.Left  + 1,  FChevronRect.Top + CHEVRON_SIZE),
               Point(FChevronRect.Right - 1,  FChevronRect.Top),
               Point(FChevronRect.Right - 1,  FChevronRect.Top + CHEVRON_SIZE * 2),
               Point(FChevronRect.Left  + 1,  FChevronRect.Top + CHEVRON_SIZE)]);

      Polygon([Point(FChevronRect.Left  + 1,  FChevronRect.Bottom - CHEVRON_SIZE),
               Point(FChevronRect.Right - 1,  FChevronRect.Bottom),
               Point(FChevronRect.Right - 1,  FChevronRect.Bottom - CHEVRON_SIZE * 2),
               Point(FChevronRect.Left  + 1,  FChevronRect.Bottom - CHEVRON_SIZE)]);

      Pen.Color   :=  clBlack;
      MoveTo(FChevronRect.Left + 3, FChevronRect.Top    + CHEVRON_SIZE * 2 + 3);
      LineTo(FChevronRect.Left + 3, FChevronRect.Bottom - CHEVRON_SIZE * 2 - 3);
      Pen.Color   :=  clWhite;
      MoveTo(FChevronRect.Left + 4, FChevronRect.Top    + CHEVRON_SIZE * 2 + 3);
      LineTo(FChevronRect.Left + 4, FChevronRect.Bottom - CHEVRON_SIZE * 2 - 3);
    end else
    if  (Align = alLeft)  and (not FOpen) or
        (Align = alRight) and (FOpen) then
    begin
      Polygon([Point(FChevronRect.Right + 1,  FChevronRect.Top + CHEVRON_SIZE),
               Point(FChevronRect.Left  - 1,  FChevronRect.Top),
               Point(FChevronRect.Left  - 1,  FChevronRect.Top + CHEVRON_SIZE * 2),
               Point(FChevronRect.Right + 1,  FChevronRect.Top + CHEVRON_SIZE)]);

      Polygon([Point(FChevronRect.Right + 1,  FChevronRect.Bottom - CHEVRON_SIZE),
               Point(FChevronRect.Left  - 1,  FChevronRect.Bottom),
               Point(FChevronRect.Left  - 1,  FChevronRect.Bottom - CHEVRON_SIZE * 2),
               Point(FChevronRect.Right + 1,  FChevronRect.Bottom - CHEVRON_SIZE)]);

      Pen.Color   :=  clBlack;
      MoveTo(FChevronRect.Left + 3, FChevronRect.Top    + CHEVRON_SIZE * 2 + 3);
      LineTo(FChevronRect.Left + 3, FChevronRect.Bottom - CHEVRON_SIZE * 2 - 3);
      Pen.Color   :=  clWhite;
      MoveTo(FChevronRect.Left + 4, FChevronRect.Top    + CHEVRON_SIZE * 2 + 3);
      LineTo(FChevronRect.Left + 4, FChevronRect.Bottom - CHEVRON_SIZE * 2 - 3);
    end else
    if  (Align = alTop)  and (FOpen) or
        (Align = alBottom) and (not FOpen) then
    begin
      Polygon([Point(FChevronRect.Left,  FChevronRect.Bottom - 1),
               Point(FChevronRect.Left + CHEVRON_SIZE,  FChevronRect.Top + 1),
               Point(FChevronRect.Left + CHEVRON_SIZE * 2,  FChevronRect.Bottom - 1),
               Point(FChevronRect.Left,  FChevronRect.Bottom - 1)]);

      Polygon([Point(FChevronRect.Right - CHEVRON_SIZE * 2,  FChevronRect.Bottom - 1),
               Point(FChevronRect.Right - CHEVRON_SIZE,  FChevronRect.Top + 1),
               Point(FChevronRect.Right,  FChevronRect.Bottom - 1),
               Point(FChevronRect.Right - CHEVRON_SIZE * 2,  FChevronRect.Bottom - 1)]);

      Pen.Color   :=  clBlack;
      MoveTo(FChevronRect.Left    + CHEVRON_SIZE * 2 + 3, FChevronRect.Top + 2);
      LineTo(FChevronRect.Right - CHEVRON_SIZE * 2 - 3,  FChevronRect.Top + 2);
      Pen.Color   :=  clWhite;
      MoveTo(FChevronRect.Left    + CHEVRON_SIZE * 2 + 3, FChevronRect.Top + 3);
      LineTo(FChevronRect.Right - CHEVRON_SIZE * 2 - 3,  FChevronRect.Top + 3);
    end else
    if  (Align = alTop)  and (not FOpen) or
        (Align = alBottom) and (FOpen) then
    begin
      Polygon([Point(FChevronRect.Left,  FChevronRect.Top + 1),
               Point(FChevronRect.Left + CHEVRON_SIZE,  FChevronRect.Bottom - 1),
               Point(FChevronRect.Left + CHEVRON_SIZE * 2,  FChevronRect.Top + 1),
               Point(FChevronRect.Left,  FChevronRect.Top + 1)]);

      Polygon([Point(FChevronRect.Right - CHEVRON_SIZE * 2,  FChevronRect.Top + 1),
               Point(FChevronRect.Right - CHEVRON_SIZE,  FChevronRect.Bottom - 1),
               Point(FChevronRect.Right,  FChevronRect.Top + 1),
               Point(FChevronRect.Right - CHEVRON_SIZE * 2,  FChevronRect.Top + 1)]);

      Pen.Color   :=  clBlack;
      MoveTo(FChevronRect.Left    + CHEVRON_SIZE * 2 + 3, FChevronRect.Top + 2);
      LineTo(FChevronRect.Right - CHEVRON_SIZE * 2 - 3,  FChevronRect.Top + 2);
      Pen.Color   :=  clWhite;
      MoveTo(FChevronRect.Left    + CHEVRON_SIZE * 2 + 3, FChevronRect.Top + 3);
      LineTo(FChevronRect.Right - CHEVRON_SIZE * 2 - 3,  FChevronRect.Top + 3);
    end;

  end;
end;

procedure TZonuSplitter.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
  MouseInChevron  := PtInRect(FChevronRect, Point(X, Y));
end;

procedure TZonuSplitter.CMMouseLeave(var Msg: TMessage);
begin
  inherited;
  MouseInChevron  := False;
end;

procedure TZonuSplitter.SetMouseInChevron(const Value: boolean);
begin
  if  (FMouseInChevron <> Value)  then  begin
    FMouseInChevron := Value;
    if  (FMouseInChevron) then  begin
      if Cursor <> FChevronCursor then begin
        FDefCursor := Cursor;
        Cursor := FChevronCursor;
      end;
    end else begin
      Cursor := FDefCursor;
    end;
    Invalidate;
  end;
end;

procedure TZonuSplitter.Click;
var
  Pos: TPoint;
  Control: TControl;
begin
  GetCursorPos(Pos);
  Pos := ScreenToClient(Pos);
  if PtInRect(FChevronRect, Pos) then begin
    Control := TDummySplitter(Self).FControl;
    if Control <> nil then  begin
      if ((Align in [alLeft, alRight]) and (Control.Width <> MinSize))
        or ((Align in [alTop, alBottom]) and (Control.Height <> MinSize)) then begin
        FOldPosition := TDummySplitter(Self).FNewSize;
        TDummySplitter(Self).FOldSize := -1;
        TDummySplitter(Self).FNewSize := MinSize;
        if  (Assigned(FOnClose))  then  begin
          FOnClose(Self);
        end;
        FOpen   :=  False;
      end else begin
        TDummySplitter(Self).FNewSize := FOldPosition;
        if  (Assigned(FOnOpen))  then  begin
          FOnOpen(Self);
        end;
        FOldPosition := TDummySplitter(Self).FNewSize;
        FOpen   :=  True;
      end;
    end;
  end;
  inherited Click;
end;

procedure TZonuSplitter.ClosePane;
begin
  ChangePane(False);
end;

procedure TZonuSplitter.OpenPane;
begin
  ChangePane(True);
end;

procedure TZonuSplitter.ChangePane(const Open: boolean);
var
  Control: TControl;
  a : integer;
begin
  MouseDown(mbLeft, [], 0, 0);
  Control := TDummySplitter(Self).FControl;
  if Control <> nil then  begin
    if ((Align in [alLeft, alRight]) and (Control.Width <> MinSize))
      or ((Align in [alTop, alBottom]) and (Control.Height <> MinSize)) then begin
      if  (not Open)  then  begin
        FOldPosition := TDummySplitter(Self).FNewSize;
        TDummySplitter(Self).FOldSize := -1;
        TDummySplitter(Self).FNewSize := MinSize;
      end;
      FOpen   :=  True;
    end else begin
      if  (Open)  then  begin
        a :=  TDummySplitter(Self).FNewSize;
        TDummySplitter(Self).FNewSize := FOldPosition;
        FOldPosition := a;
      end;
      FOpen   :=  False;
    end;
  end;
  MouseUp(mbLeft, [], 0, 0);
  StopSizing;
end;

procedure TZonuSplitter.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited;
  if  (TDummySplitter(Self).FNewSize > MinSize) and (not FOpen)  then  begin
    if  (Assigned(FOnOpen)) then  begin
      FOnOpen(Self);
    end;
    FOpen :=  True;
  end;
  if  (TDummySplitter(Self).FNewSize <= MinSize) and (FOpen) then begin //(FOldPosition > MinSize)  then  begin
    if  (Assigned(FOnClose)) then  begin
      FOnClose(Self);
    end;
    FOpen :=  False;
  end;
end;

end.
