unit ExifInfo;  // ExifInfo ver.2.5

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Jpeg, Math;

type
  TDatas = packed record
    FloatValue : Double;
    SignedInt  : Integer;
    Text       : string;
    UnsignedInt: Cardinal;
  end;
  TExifOrientation = (eoNone,       eoTopLeft, eoTopRight, eoBottomRight,
                      eoBottomLeft, eoLeftTop, eoRightTop, eoRightBottom,
                      eoLeftBottom);
  TExifRational = packed record
    RationalText: string;
    Value       : Double;
  end;
  TExifStringValue = packed record
    Text : string;
    Value: Integer;
  end;
  TWriteType = (wtII, wtMM);
  TIFDPosSet = packed record
    IFD0Sub: Int64;
    IFD0Int: Int64;
    IFD1   : Int64;
    IFD1Sub: Int64;
    IFD1Int: Int64;
  end;

  // #################################### CR|[lg[ExifInfo^] #####
  TExifInfo = class(TComponent)
  private
    { Private 錾 }
    // ================================================== tB[h ===
    FWriteType: TWriteType;
    PicFile: TFileStream;
    // ============================================ vpeBtB[h ===
    FApertureValue: TExifRational;
    FBrightness: TExifRational;
    FColorSpace: TExifStringValue;
    FCompression: TExifStringValue;
    FCompressBitsPerPixel: TExifRational;
    FCopyright: string;
    FDateTime: string;
    FDateTimeDigitized: string;
    FDateTimeOriginal: string;
    FExifVersion: string;
    FExposureBias: TExifRational;
    FExposureProgram: TExifStringValue;
    FExposureTime: TExifRational;
    FFlash: TExifStringValue;
    FFlashPixVersion: string;
    FFNumber: TExifRational;
    FFocalLength: TExifRational;
    FFocalPlaneResolutionUnit: TExifStringValue;
    FFocalPlaneXResolution: TExifRational;
    FFocalPlaneYResolution: TExifRational;
    FImageDescription: string;
    FImageFileName: string;
    FImageHeight: Integer;
    FImageWidth: Integer;
    FISOSpeedRating: Word;
    FJpegInterchangeFormat: Cardinal;
    FJpegInterchangeLength: Cardinal;
    FLightSource: TExifStringValue;
    FMake: string;
    FMakerNoteLength: Cardinal;
    FMakerNotePosition: Cardinal;
    FMaxAperture: TExifRational;
    FMeteringMode: TExifStringValue;
    FModel: string;
    FOrientation: TExifOrientation;
    FOnReadError: TNotifyEvent;
    FPhotometricInterpretation: TExifStringValue;
    FPlanarConfiguration: Word;
    FRelatedSoundFile: string;
    FResolutionUnit: TExifStringValue;
    FSamplesPerPixel: Word;
    FSearchRange: Integer;
    FSensingMethod: TExifStringValue;
    FShutterSpeed: TExifRational;
    FSoftware: string;
    FSubjectDistance: TExifRational;
    FSubSecTime: string;
    FSubSecTimeDigitized: string;
    FSubSecTimeOriginal: string;
    FThmImageHeight: Integer;
    FThmImageWidth: Integer;
    FThmOrientation: TExifOrientation;
    FThmResolutionUnit: TExifStringValue;
    FThmXResolution: TExifRational;
    FThmYResolution: TExifRational;
    FUserComment: TStringList;
    FXResolution: TExifRational;
    FYResolution: TExifRational;
    // ============================================ vCx[g\bh ===
    procedure ClearFields;
    function ExtractUserComment(const HexText: string): string;
    procedure GetExifData(ImgFile: TStream; var ExifData: TDatas;
        const DataType: ShortInt; const DataCount: Longword;
        const TiffTop: Int64);
    procedure GetExifValues(const TiffTop: Int64);
    function SignedMakeTag(const Value; const AByte: Byte): Integer;
    function SrcExifHeader(const FileName: string;
        var Position: Int64): Boolean;
    function UndefToText(const Undef: string): string;
    function UnsignedMakeTag(const Value; const AByte: Byte): Cardinal;
    // ============================================== vpeBANZX ===
    procedure SetImageFileName(const Value: string);
    function GetUserComment: TStrings;
  protected
    { Protected 錾 }
    // ================================================ Cxg\bh ===
    procedure ReadError; dynamic;
  public
    { Public 錾 }
    // ============================================== vpeB(public) ===
    property ApertureValue: TExifRational read FApertureValue;
    property Brightness: TExifRational read FBrightness;
    property ColorSpace: TExifStringValue read FColorSpace;
    property Compression: TExifStringValue read FCompression;
    property CompressBitsPerPixel: TExifRational read FCompressBitsPerPixel;
    property Copyright: string read FCopyright;
    property DateTime: string read FDateTime;
    property DateTimeDigitized: string read FDateTimeDigitized;
    property DateTimeOriginal: string read FDateTimeOriginal;
    property ExifVersion: string read FExifVersion;
    property ExposureBias: TExifRational read FExposureBias;
    property ExposureProgram: TExifStringValue read FExposureProgram;
    property ExposureTime: TExifRational read FExposureTime;
    property Flash: TExifStringValue read FFlash;
    property FlashPixVersion: string read FFlashPixVersion;
    property FNumber: TExifRational read FFNumber;
    property FocalLength: TExifRational read FFocalLength;
    property FocalPlaneResolutionUnit: TExifStringValue
        read FFocalPlaneResolutionUnit;
    property FocalPlaneXResolution: TExifRational read FFocalPlaneXResolution;
    property FocalPlaneYResolution: TExifRational read FFocalPlaneYResolution;
    property ImageDescription: string read FImageDescription;
    property ImageHeight: Integer read FImageHeight;
    property ImageWidth: Integer read FImageWidth;
    property ISOSpeedRating: Word read FISOSpeedRating;
    property JpegInterchangeFormat: Cardinal read FJpegInterchangeFormat;
    property JpegInterchangeLength: Cardinal read FJpegInterchangeLength;
    property LightSource: TExifStringValue read FLightSource;
    property Make: string read FMake;
    property MakerNoteLength: Cardinal read FMakerNoteLength;
    property MakerNotePosition: Cardinal read FMakerNotePosition;
    property MaxAperture: TExifRational read FMaxAperture;
    property MeteringMode: TExifStringValue read FMeteringMode;
    property Model: string read FModel;
    property Orientation: TExifOrientation read FOrientation;
    property PhotometricInterpretation: TExifStringValue
        read FPhotometricInterpretation;
    property PlanarConfiguration: Word read FPlanarConfiguration;
    property RelatedSoundFile: string read FRelatedSoundFile;
    property ResolutionUnit: TExifStringValue read FResolutionUnit;
    property SamplesPerPixel: Word read FSamplesPerPixel;
    property SensingMethod: TExifStringValue read FSensingMethod;
    property ShutterSpeed: TExifRational read FShutterSpeed;
    property Software: string read FSoftware;
    property SubjectDistance: TExifRational read FSubjectDistance;
    property SubSecTime: string read FSubSecTime;
    property SubSecTimeDigitized: string read FSubSecTimeDigitized;
    property SubSecTimeOriginal: string read FSubSecTimeOriginal;
    property ThmImageHeight: Integer read FThmImageHeight;
    property ThmImageWidth: Integer read FThmImageWidth;
    property ThmOrientation: TExifOrientation read FThmOrientation;
    property ThmResolutionUnit: TExifStringValue read FThmResolutionUnit;
    property ThmXResolution: TExifRational read FThmXResolution;
    property ThmYResolution: TExifRational read FThmYResolution;
    property UserComment: TStrings read GetUserComment;
    property XResolution: TExifRational read FXResolution;
    property YResolution: TExifRational read FYResolution;
    // ======================================================== \bh ===
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure CopyExif(const Dest: string);
    function ExifCheck(const FileName: string): Boolean;
    function GetThumbnail(Thumb: TBitmap): Boolean;
  published
    { Published 錾 }
    // ====================================================== vpeB ===
    property ImageFileName: string read FImageFileName write SetImageFileName;
    property SearchRange: Integer read FSearchRange write FSearchRange
        default 32;
    // ======================================================== Cxg ===
    property OnReadError: TNotifyEvent read FOnReadError write FOnReadError;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('NonVCL', [TExifInfo]);
end;

const
  EMSG_OPEN = '%s ݂͑ȂA܂͊J܂B';
  EMSG_NOTJPEG = '%s ͖ȃt@CłB';

// ############################################################# TExifInfo #####
{ TExifInfo }

// ************************** g[^Ce`𔻕ʂ16iɕϊ
function TExifInfo.SignedMakeTag(const Value; const AByte: Byte): Integer;
var
  Lp: Integer;
  Lw: Longint;
  S, S1: string;
  W: SmallInt;
begin
  Result := 0;
  if AByte = 2 then
  begin
    W := SmallInt(Value);
    case FWriteType of
      wtII: Result := W;
      wtMM:
        begin
          S := IntToHex(Lo(W), 2) + IntToHex(Hi(W), 2);
          Result := StrToInt('$' + S);
        end;
    end;
  end
  else
  begin
    Lw := Longint(Value);
    case FWriteType of
      wtII: Result := Lw;
      wtMM:
      begin
        S := IntToHex(Lw, AByte * 2);
        S1 := '';
        for Lp := 4 downto 1 do
          S1 := S1 + Copy(S, Lp * 2 - 1, 2);
        Result := StrToInt('$' + S1);
      end;
    end;
  end;
end;

// ################################## TlC擾<GetThumbnail\bh> #####
function TExifInfo.GetThumbnail(Thumb: TBitmap): Boolean;
var
  Pic: TFileStream;  // 摜t@C
  Thm: TJpegImage;   // TlC
  ThmPos: Int64;     // TlC̊i[ʒu
  TiffPos: Int64;    // TIFFf[^gbvʒu
begin
  // ************************************** TIFFf[^̃gbvʒu擾 ***
  Result := SrcExifHeader(FImageFileName, TiffPos);
  if not Result then Exit;
  // **************************************************** L^ʒůmF ***
  Result := (FJpegInterchangeFormat >= 1) and (FJpegInterchangeLength >= 1);
  if not Result then Exit;
  ThmPos := TiffPos + FJpegInterchangeFormat;
  // ******************************************** L^tH[}bg̊mF ***
  Result := (FCompression.Value = 6);
  if not Result then Exit;
  // **************************************************** TlC擾 ***
  try
    Pic := TFileStream.Create(FImageFileName, fmOpenRead or fmShareDenyWrite);
    try
      Result := (Pic.Seek(ThmPos, soBeginning) = ThmPos);
      if Result then
      begin
        Thm := TJpegImage.Create;
        try
          Thm.LoadFromStream(Pic);
          Thumb.Assign(Thm);
        finally
          Thm.Free;
        end;
      end;
    finally
      Pic.Free;
    end;
  except
    on Exception do Result := False;
  end;
end;

// ############################### Exifwb_̃`FbN<ExifCheck\bh> #####
function TExifInfo.ExifCheck(const FileName: string): Boolean;
var
  Position: Int64;
begin
  Result := SrcExifHeader(FileName, Position);
end;

// ############### Exifwb_TÄʒuԂ<SrcExifHeader\bh> #####
function TExifInfo.SrcExifHeader(const FileName: string;
  var Position: Int64): Boolean;
var
  Buf2: Word;
  Rs: Integer;                    // ǂݍ񂾃f[^̃TCY
  ExifStr: array[0..6] of Char;
begin
  Result := False;
  if (FileName <> '') and FileExists(FileName) then
  try
    // ************************************************** wb_̃T[` ***
    PicFile := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
    try
      if PicFile.Read(Buf2, 2) = 2 then
      begin
        // ============================================ JPEGwb_̊mF ===
        if Buf2 = $D8FF then
        begin
          // ================================= APP1($FFE1)}[J[̌ ===
          PicFile.Seek(1, soCurrent);
          repeat
            PicFile.Seek(-1, soCurrent);
            Rs := PicFile.Read(Buf2, 2);
            Result := (Rs = 2);
          until (Buf2 = $E1FF) or (not Result) or (PicFile.Position > FSearchRange);
          // ================ ^OĂExifwb_ǂݎ ===
          if Buf2 = $E1FF then
          begin
            PicFile.Seek(2, soCurrent);
            PicFile.Read(ExifStr, 6);
            if SameText(ExifStr, 'exif') then
              Position := PicFile.Position
            else
              Result := False;
          end
          else
            Result := False;
        end;
      end;
    finally
      PicFile.Free;
    end;
  except
    Result := False;
  end;
end;

// ######### ImageFileNamevpeBݒ<SetImageFileNameANZX\bh> #####
procedure TExifInfo.SetImageFileName(const Value: string);
var
  TiffHeadPos: Int64;   // TIFFwb_̊Jnʒu
begin
  FImageFileName := Value;
  // ************************************************ tB[h̏ ***
  ClearFields;
  // ************************************************************** 擾 ***
  if SrcExifHeader(FImageFileName, TiffHeadPos) then
    GetExifValues(TiffHeadPos)
  else
    ReadError;
end;

// ################### Exifef[^ǂݍ<GetExifValues\bh> #####
procedure TExifInfo.GetExifValues(const TiffTop: Int64);
var
  BufLong: Longword;        // LongWord^obt@
  BufWord: Word;            // Word^obt@
  CurrentIfd: Integer;      // ݎ擾IFDGgԍ
  DataCount: Longword;      // f[^̌
  DataType: ShortInt;       // f[^̊i[`
  DirEntry: Word;           // fBNgGg
  ExifData: TDatas;         // 擾f[^
  IFDCount: Integer;
  IFDEntry: array[0..12] of Int64; // IFD̃AhXXg
  Interoperability: Boolean;  // TlC摜IFD
  LDir: Word;
  Offset: Longword;         // ItZbg
  Ps: TFileStream;          // 摜t@C
  Res: Integer;             // 擾TCY
  Tag: Integer;             // ^O
begin
  Ps := TFileStream.Create(FImageFileName, fmOpenRead or fmShareDenyWrite);
  try
    // ******************************************** TIFFwb_̓ǂݎ ***
    Ps.Seek(TiffTop, soBeginning);
    // ================================================== f[^i[` ===
    Ps.ReadBuffer(BufWord, 2);
    if BufWord = $4D4D then FWriteType := wtMM else FWriteType := wtII;
    // ================================================== ItZbg擾 ===
    Ps.Seek(2, soCurrent);
    Ps.Read(Offset, 4);
    Offset := UnsignedMakeTag(Offset, 4);
    // ******************************************************* IFD̉ ***
    CurrentIfd := 0;
    IFDEntry[0] := Ps.Position + Offset - 8;
    IFDCount := 0;
    FMakerNotePosition := 0;
    Interoperability := False;
    // ============================================= IFD̐JԂ ===
    repeat
      Ps.Seek(IFDEntry[CurrentIfd], soBeginning);
      // ---------------------------------- fBNgGg擾 ---
      Res := Ps.Read(DirEntry, 2);
      DirEntry := UnsignedMakeTag(DirEntry, 2);
      if Res <> 2 then Break;
      // ------------------------------------------ GǧJԂ ---
      for LDir := 1 to DirEntry do
      begin
        // .................................................. ^O̎擾 ...
        Ps.Read(BufWord, 2);
        Tag := UnsignedMakeTag(BufWord, 2);
        // ................................................ f[^̎擾 ...
        Ps.Read(BufWord, 2);
        DataType := SignedMakeTag(BufWord, 2);
        Ps.Read(BufLong, 4);
        DataCount := UnsignedMakeTag(BufLong, 4);
        GetExifData(Ps, ExifData, DataType, DataCount, TiffTop);
        Ps.Seek(4, soCurrent);
        // .................................................. ^O̔ ...
        case Tag of
        $0001:  // InteroperabilityIndex
          Interoperability := SameText(ExifData.Text, 'THM');
        $0100:  // ImageWidth
          FThmImageWidth := ExifData.UnsignedInt;
        $0101:  // ImageHeight
          FThmImageHeight := ExifData.UnsignedInt;
        $0103:  // Compression
          begin
            FCompression.Value := ExifData.UnsignedInt;
            case ExifData.UnsignedInt of
              1: FCompression.Text := '񈳏k';
              6: FCompression.Text := 'JPEG';
              else
                FCompression.Text := '';
            end;
          end;
        $0106:  // PhotometricInterpretation
          with FPhotometricInterpretation do
          begin
            Value := ExifData.UnsignedInt;
            case Value of
              2: Text := 'RGB';
              6: Text := 'YCbCr';
              else
                Text := '';
            end;
          end;
        $010E:  // ImageDiscription
          FImageDescription := ExifData.Text;
        $010F:  // Make
          FMake := ExifData.Text;
        $0110:  // Model
          FModel := ExifData.Text;
        $0112:  // Orientation
          if CurrentIfd = 0 then
          case ExifData.UnsignedInt of
            0: FOrientation := eoNone;
            1: FOrientation := eoTopLeft;
            2: FOrientation := eoTopRight;
            3: FOrientation := eoBottomLeft;
            4: FOrientation := eoBottomRight;
            5: FOrientation := eoLeftTop;
            6: FOrientation := eoRightTop;
            7: FOrientation := eoRightBottom;
            8: FOrientation := eoLeftBottom;
          end
          else
          case ExifData.UnsignedInt of
            0: FThmOrientation := eoNone;
            1: FThmOrientation := eoTopLeft;
            2: FThmOrientation := eoTopRight;
            3: FThmOrientation := eoBottomLeft;
            4: FThmOrientation := eoBottomRight;
            5: FThmOrientation := eoLeftTop;
            6: FThmOrientation := eoRightTop;
            7: FThmOrientation := eoRightBottom;
            8: FThmOrientation := eoLeftBottom;
          end;
        $0115:  // SamplesPerPixel
          FSamplesPerPixel := ExifData.UnsignedInt;
        $011A:  // XResolution
          if CurrentIfd = 0 then
          begin
            FXResolution.RationalText := ExifData.Text;
            FXResolution.Value := ExifData.FloatValue;
          end
          else
          begin
            FThmXResolution.RationalText := ExifData.Text;
            FThmXResolution.Value := ExifData.FloatValue;
          end;
        $011B:  // YResolution
          if CurrentIfd = 0 then
          begin
            FYResolution.RationalText := ExifData.Text;
            FYResolution.Value := ExifData.FloatValue;
          end
          else
          begin
            FThmYResolution.RationalText := ExifData.Text;
            FThmYResolution.Value := ExifData.FloatValue;
          end;
        $011C:  // PlanarConfiguration
          FPlanarConfiguration := ExifData.UnsignedInt;
        $0128:  // ResolutionUnit
          if CurrentIfd = 0 then
          begin
            FResolutionUnit.Value := ExifData.UnsignedInt;
            case ExifData.UnsignedInt of
              2: FResolutionUnit.Text := 'inch';
              3: FResolutionUnit.Text := 'cm';
            end;
          end
          else
          begin
            FThmResolutionUnit.Value := ExifData.UnsignedInt;
            case ExifData.UnsignedInt of
              2: FThmResolutionUnit.Text := 'inch';
              3: FThmResolutionUnit.Text := 'cm';
            end;
          end;
        $0131:  // Software
          FSoftware := ExifData.Text;
        $0132:  // DateTime
          FDateTime := ExifData.Text;
        $0201:  // JpegInterchangeFormat
          FJpegInterchangeFormat := ExifData.UnsignedInt;
        $0202:  // JpegInterchangeFormatLength
          FJpegInterchangeLength := ExifData.UnsignedInt;
        $1001:  // RelatedImageWidth
          if Interoperability then
          begin
            if FThmImageWidth = 0 then FThmImageWidth := ExifData.UnsignedInt;
          end
          else
            if FImageWidth = 0 then FImageWidth := ExifData.UnsignedInt;
        $1002:  // RelatedImageHeight
          if Interoperability then
          begin
            if FThmImageHeight = 0 then FThmImageHeight := ExifData.UnsignedInt;
          end
          else
            if FImageHeight = 0 then FImageHeight := ExifData.UnsignedInt;
        $8298:  // Copyright
          FCopyright := ExifData.Text;
        $829A:  // ExposureTime
          begin
            FExposureTime.RationalText := ExifData.Text;
            FExposureTime.Value := ExifData.FloatValue;
          end;
        $829D:  // FNumber
          begin
            FFNumber.RationalText := ExifData.Text;
            FFNumber.Value := ExifData.FloatValue;
          end;
        $8769:  // Exif Sub IFDւ̃ItZbg
          begin
            Inc(IFDCount);
            IFDEntry[IFDCount] := TiffTop + ExifData.UnsignedInt;
          end;
        $8822:  // ExposureProgram
          begin
            FExposureProgram.Value := ExifData.UnsignedInt;
            case ExifData.UnsignedInt of
              1: FExposureProgram.Text := '}jAݒ';
              2: FExposureProgram.Text := 'vOAE';
              3: FExposureProgram.Text := 'iD';
              4: FExposureProgram.Text := 'Vb^[xD';
              5: FExposureProgram.Text := 'ᑬvO';
              6: FExposureProgram.Text := 'vO';
              7: FExposureProgram.Text := '|[g[g[h';
              8: FExposureProgram.Text := 'i[h';
              else
                FExposureProgram.Text := '';
            end;
          end;
        $8827:  // ISOSpeedRating
          FISOSpeedRating := ExifData.UnsignedInt;
        $9000:  // Exif Version
          FExifVersion := UndefToText(ExifData.Text);
        $9003:  // DateTimeOriginal
          FDateTimeOriginal := ExifData.Text;
        $9004:  // DateTimeDigitized
          FDateTimeDigitized := ExifData.Text;
        $9102:  // CompressBitsPerPixel
          begin
            FCompressBitsPerPixel.RationalText := ExifData.Text;
            FCompressBitsPerPixel.Value := ExifData.FloatValue;
          end;
        $9201:  // ShutterSpeedValue
          begin
            FShutterSpeed.RationalText := ExifData.Text;
            FShutterSpeed.Value := ExifData.FloatValue;
          end;
        $9202:  // ApertureValue
          begin
            FApertureValue.RationalText := ExifData.Text;
            FApertureValue.Value := ExifData.FloatValue;
          end;
        $9203:  // BrightnessValue
          begin
            FBrightness.RationalText := ExifData.Text;
            FBrightness.Value := ExifData.FloatValue;
          end;
        $9204:  // ExposureBiasValue
          begin
            FExposureBias.RationalText := ExifData.Text;
            FExposureBias.Value := ExifData.FloatValue;
          end;
        $9205:  // MaxApertureValue
          begin
            FMaxAperture.RationalText := ExifData.Text;
            FMaxAperture.Value := ExifData.FloatValue;
          end;
        $9206:  // SubjectDistance
          begin
            FSubjectDistance.RationalText := ExifData.Text;
            FSubjectDistance.Value := ExifData.FloatValue;
          end;
        $9207:  // MeteringMode
          begin
            FMeteringMode.Value := ExifData.UnsignedInt;
            case ExifData.UnsignedInt of
              0: FMeteringMode.Text := 's';
              1: FMeteringMode.Text := 'ϑ';
              2: FMeteringMode.Text := 'd_';
              3: FMeteringMode.Text := 'X|bg';
              4: FMeteringMode.Text := '_X|bg';
              5: FMeteringMode.Text := '}`ZOg';
              6: FMeteringMode.Text := '';
              255: FMeteringMode.Text := '̑';
              else
                FMeteringMode.Text := '';
            end;
          end;
        $9208:  // LightSource
          begin
            FLightSource.Value := ExifData.UnsignedInt;
            case ExifData.UnsignedInt of
              0: FLightSource.Text := 's';
              1: FLightSource.Text := '';
              2: FLightSource.Text := 'u';
              3: FLightSource.Text := 'Md';
              10: FLightSource.Text := 'tbV';
              17: FLightSource.Text := 'WCgA';
              18: FLightSource.Text := 'WCgB';
              19: FLightSource.Text := 'WCgC';
              20: FLightSource.Text := 'D55';
              21: FLightSource.Text := 'D65';
              22: FLightSource.Text := 'D75';
              255: FLightSource.Text := '̑';
              else
                FLightSource.Text := '';
            end;
          end;
        $9209:  // Flash
          begin
            FFlash.Value := ExifData.UnsignedInt;
            case ExifData.UnsignedInt of
              0: FFlash.Text := '񔭌';
              1: FFlash.Text := '(ˌo@\Ȃ)';
              5: FFlash.Text := '(ˌo)';
              7: FFlash.Text := '(ˌo)';
              else
                FFlash.Text := '';
            end;
          end;
        $920A:  // FocalLength
          begin
            FFocalLength.RationalText := ExifData.Text;
            FFocalLength.Value := ExifData.FloatValue;
          end;
        $927C:  // MakerNote
          begin
            Ps.Seek(-4, soCurrent);
            Ps.Read(BufLong, 4);
            FMakerNoteLength := DataCount;
            FMakerNotePosition := UnsignedMakeTag(BufLong, 4) + TiffTop;
          end;
        $9286:  // UserComment
          if DataType <> 7 then
            case DataType of
              1, 3, 4 : FUserComment.Append(IntToStr(ExifData.UnsignedInt));
              2, 5, 10: FUserComment.Append(ExifData.Text);
              6, 8, 9 : FUserComment.Append(IntToStr(ExifData.SignedInt));
              11, 12  : FUserComment.Append(FloatToStr(ExifData.FloatValue));
            end
          else
          begin
            if ExifData.Text <> '' then
              FUserComment.Append(ExtractUserComment(ExifData.Text));
          end;
        $9290:  // SubSecTime
          FSubSecTime := ExifData.Text;
        $9291:  // SubSecTimeOriginal
          FSubSecTimeOriginal := ExifData.Text;
        $9292:  // SubSecTimeDigitized
          FSubSecTimeDigitized := ExifData.Text;
        $A000:  // FlashPixVersion
          FFlashPixVersion := UndefToText(ExifData.Text);
        $A001:  // ColorSpace
          begin
            FColorSpace.Value := ExifData.UnsignedInt;
            case ExifData.UnsignedInt of
              1    : FColorSpace.Text := 'sRGB';
              65535: FColorSpace.Text := 'Uncalibrated';
              else
                FColorSpace.Text := '';
            end;
          end;
        $A002:  // ExifImageWidth
          FImageWidth := ExifData.UnsignedInt;
        $A003:  // ExifImageHeight
          FImageHeight := ExifData.UnsignedInt;
        $A004:  // RelatedSoundFile
          FRelatedSoundFile := ExifData.Text;
        $A005:  // InteroperabilityIFDPointer
          begin
            Inc(IFDCount);
            IFDEntry[IFDCount] := TiffTop + ExifData.UnsignedInt;
          end;
        $A20E:  // FocalPlaneXResolution
          begin
            FFocalPlaneXResolution.RationalText := ExifData.Text;
            FFocalPlaneXResolution.Value := ExifData.FloatValue;
          end;
        $A20F:  // FocalPlaneYResolution
          begin
            FFocalPlaneYResolution.RationalText := ExifData.Text;
            FFocalPlaneYResolution.Value := ExifData.FloatValue;
          end;
        $A210:  // FocalPlaneResolutionUnit
          begin
            FFocalPlaneResolutionUnit.Value := ExifData.UnsignedInt;
            case ExifData.UnsignedInt of
              2: FFocalPlaneResolutionUnit.Text := 'inch';
              3: FFocalPlaneResolutionUnit.Text := 'cm';
              else
                FFocalPlaneResolutionUnit.Text := '';
            end;
          end;
        $A215:  // ExposureIndex
          FISOSpeedRating := Round(ExifData.FloatValue);
        $A217:  // SensingMethod
          begin
            FSensingMethod.Value := ExifData.UnsignedInt;
            case ExifData.UnsignedInt of
              2: FSensingMethod.Text := '1`bvJ[GAZT';
              3: FSensingMethod.Text := '2`bvJ[GAZT';
              4: FSensingMethod.Text := '3`bvJ[GAZT';
              5: FSensingMethod.Text := 'J[V[NGVGAZT';
              7: FSensingMethod.Text := 'gjAZT';
              8: FSensingMethod.Text := 'J[V[NGVjAZT';
              else
                FSensingMethod.Text := '';
            end;
          end;
        end;
      end;
      Ps.Read(Offset, 4);
      if Offset >= 1 then
      begin
        Inc(IFDCount);
        IFDEntry[IFDCount] := Offset + TiffTop;
      end;
      // ----------------------------------------------------- IFD ---
      Inc(CurrentIfd);
    until CurrentIfd > IFDCount;
  finally
    Ps.Free;
  end;
end;

// ################## OnReadErrorCxg̔<ReadErrorCxg\bh> #####
procedure TExifInfo.ReadError;
begin
  if Assigned(FOnReadError) then FOnReadError(Self);
end;

// ################################# Exiff[^̎擾<GetExifData\bh> #####
procedure TExifInfo.GetExifData(ImgFile: TStream; var ExifData: TDatas;
    const DataType: ShortInt; const DataCount: Longword; const TiffTop: Int64);
var
  AllBytes: Cardinal;  // f[^̑oCg
  Offset: Longword;    // ItZbg
  TopPos: Int64;       // 擾Jnʒu
  // ++++++++++++++++++++++++++ ۂ̃f[^ǂݍݎ葱(ReadData葱) +++++
  procedure ReadData;
  var
    BufChar: Char;            // Char^obt@
    BufByte: Byte;            // Byte^obt@
    BufInt64: Int64;          // Int64^obt@
    BufLong: Longword;        // Longword^obt@
    BufShort: ShortInt;       // ShortInt^obt@
    BufSingle: Single;        // Single^obt@
    BufSLong: Longint;        // Longint^obt@
    BufSmall: SmallInt;       // SmallInt^obt@
    BufWord: Word;            // Word^obt@
    Denomi1, Nume1: Longword; // ȂAq
    Denomi2, Nume2: Longint;  // tAq
    Lp: Integer;
    P: Pointer;
    S: string;
  begin
    case DataType of
      1:  // ============================================= unsigned byte ===
        begin
          ImgFile.Read(BufByte, 1);
          ExifData.UnsignedInt := BufByte;
        end;
      2:  // ============================================= ascii strings ===
        begin
          for Lp := 1 to DataCount do
          begin
            ImgFile.Read(BufChar, 1);
            if BufChar <> #0 then ExifData.Text := ExifData.Text + BufChar;
          end;
        end;
      3:  // ============================================ unsigned short ===
        begin
          ImgFile.Read(BufWord, 2);
          ExifData.UnsignedInt := UnsignedMakeTag(BufWord, 2);
        end;
      4:  // ============================================= unsigned long ===
        begin
          ImgFile.Read(BufLong, 4);
          ExifData.UnsignedInt := UnsignedMakeTag(BufLong, 4);
        end;
      5:  // ========================================= unsigned rational ===
        begin
          ImgFile.Read(BufLong, 4);
          Nume1 := UnsignedMakeTag(BufLong, 4);
          ImgFile.Read(BufLong, 4);
          Denomi1 := UnsignedMakeTag(BufLong, 4);
          ExifData.Text := Format('%u/%u', [Nume1, Denomi1]);
          if Denomi1 <> 0 then ExifData.FloatValue := Nume1 / Denomi1;
        end;
      6:  // =============================================== signed byte ===
        begin
          ImgFile.Read(BufShort, 1);
          ExifData.SignedInt := BufShort;
        end;
      7:  // ================================================= undefined ===
        begin
          for Lp := 1 to DataCount do
          begin
            ImgFile.Read(BufByte, 1);
            ExifData.Text := ExifData.Text + IntToHex(BufByte, 2);
          end;
        end;
      8:  // ============================================== signed short ===
        begin
          ImgFile.Read(BufSmall, 2);
          ExifData.SignedInt := SignedMakeTag(BufSmall, 2);
        end;
      9:  // =============================================== signed long ===
        begin
          ImgFile.Read(BufSLong, 4);
          ExifData.SignedInt := SignedMakeTag(BufSLong, 4);
        end;
      10: // =========================================== signed rational ===
        begin
          ImgFile.Read(BufSLong, 4);
          Nume2 := SignedMakeTag(BufSLong, 4);
          ImgFile.Read(BufSLong, 4);
          Denomi2 := SignedMakeTag(BufSLong, 4);
          ExifData.Text := Format('%d/%d', [Nume2, Denomi2]);
          if Denomi2 <> 0 then ExifData.FloatValue := Nume2 / Denomi2;
        end;
      11: // ============================================== single float ===
        begin
          // ------------------------------------------ ^œǂݎ ---
          ImgFile.Read(BufLong, 4);
          BufLong := UnsignedMakeTag(BufLong, 4);
          // ---------------------------------------------- ^ɕϊ ---
          P := @BufLong;
          BufSingle := PSingle(P)^;
          ExifData.FloatValue := BufSingle;
        end;
      12: // ============================================== double float ===
        begin
          // ------------------------------------------ ^œǂݎ ---
          ImgFile.Read(BufInt64, 8);
          // -------------------------------------------- ̓ւ ---
          if FWriteType = wtMM then
          begin
            S := IntToHex(BufInt64, 16);
            for Lp := 8 downto 1 do
              ExifData.Text := ExifData.Text + Copy(S, Lp * 2 - 1, 2);
            BufInt64 := StrToInt('$' + ExifData.Text);
            ExifData.Text := '';
          end;
          // ---------------------------------------------- ^ɕϊ ---
          P := @BufInt64;
          ExifData.FloatValue := PDouble(P)^;
        end;
    end;
  end;
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
begin
  // **************************************************** f[^̏ ***
  with ExifData do
  begin
    FloatValue := 0;
    SignedInt := 0;
    Text := '';
    UnsignedInt := 0;
  end;
  // **************************************************************  ***
  case DataType of
    1, 2, 6, 7: AllBytes := DataCount;
    3, 8      : AllBytes := DataCount * 2;
    4, 9, 11  : AllBytes := DataCount * 4;
    else
      AllBytes := DataCount * 8;
  end;
  TopPos := ImgFile.Position;
  // ************************************** ItZbg܂̓f[^̎擾 ***
  if AllBytes <= 4 then
    ReadData
  else
  begin
    ImgFile.Read(Offset, 4);
    Offset := UnsignedMakeTag(Offset, 4) + TiffTop;
    ImgFile.Seek(Offset, soBeginning);
    ReadData;
  end;
  // **************************************************** ʒuɖ߂ ***
  ImgFile.Seek(TopPos, soBeginning);
end;

// ######################### Ȃf[^̎擾<UnsigendMakeTag\bh> #####
function TExifInfo.UnsignedMakeTag(const Value;
  const AByte: Byte): Cardinal;
var
  Lp: Integer;
  Lw: Longword;
  S, S1: string;
  W: Word;
begin
  Result := 0;
  if AByte = 2 then
  begin
    W := Word(Value);
    case FWriteType of
      wtII: Result := W;
      wtMM:
        begin
          S := IntToHex(Lo(W), 2) + IntToHex(Hi(W), 2);
          Result := StrToInt('$' + S);
        end;
    end;
  end
  else
  begin
    Lw := Longword(Value);
    case FWriteType of
      wtII: Result := Lw;
      wtMM:
      begin
        S := IntToHex(Lw, AByte * 2);
        S1 := '';
        for Lp := 4 downto 1 do
          S1 := S1 + Copy(S, Lp * 2 - 1, 2);
        Result := StrToInt('$' + S1);
      end;
    end;
  end;
end;

// ############### `f[^16i𕶎ɕϊ<UndefToText\bh> #####
function TExifInfo.UndefToText(const Undef: string): string;
var
  CNo: Byte;   // R[h
  Lp: Integer;
begin
  Result := '';
  for Lp := 1 to Length(Undef) div 2 do
  begin
    CNo := StrToInt('$' + Copy(Undef, Lp * 2 - 1, 2));
    if CNo <> 0 then Result := Result + Chr(CNo);
  end;
end;

// ###################### [URg̎擾<ExtractUserComment\bh> #####
function TExifInfo.ExtractUserComment(const HexText: string): string;
var
  Code: Word;           // 2oCgR[h
  CodeType: string;     // L^R[h̎
  HiByte, LoByte: Byte; // ʃoCgAʃoCg
  Lp: Integer;
  Mem: TMemoryStream;   // R[hϊp̃Xg[
  S: string;
  ShiftIn: Boolean;     // 2oCgR[h
  Wc: WideChar;         // Unicode
begin
  // ************************************************** R[hʂ̎擾 ***
  CodeType := UndefToText(Copy(HexText, 1, 16));
  // ****************************************************** ̎擾 ***
  // ========================================================== AXL[ ===
  if SameText(CodeType, 'ASCII') then
    Result := UndefToText(Copy(HexText, 17, Length(HexText) - 16))
  // =========================================================== Unicode ===
  else if SameText(CodeType, 'Unicode') then
  begin
    S := Copy(HexText, 17, Length(HexText) - 16);
    // ------------------------------------------- 16iUnicodeɕϊ ---
    Mem := TMemoryStream.Create;
    try
      // ............................ lɕϊăXg[ɏ ...
      for Lp := 1 to Length(S) div 4 do
      begin
        if FWriteType = wtMM then
          Code := StrToInt('$' + Copy(S, Lp * 4 - 1, 2) + Copy(S, Lp * 4 - 3, 2))
        else
          Code := StrToInt('$' + Copy(S, Lp * 4 - 3, 4));
        Mem.Write(Code, 2);
      end;
      // ........................... Xg[Unicodeǂݎ ...
      Mem.Seek(0, soBeginning);
      Result := '';
      for Lp := 1 to Length(S) div 4 do
      begin
        Mem.Read(Wc, 2);
        Result := Result + Wc;
      end;
    finally
      Mem.Free;
    end;
  end
  // ========================================================= JisR[h ===
  else if SameText(CodeType, 'JIS') then
  begin
    S := Copy(HexText, 17, Length(HexText) - 16);
    // ------------------------------ 16i𐔒lɂăXg[ ---
    Mem := TMemoryStream.Create;
    try
      for Lp := 1 to Length(S) div 2 do
      begin
        HiByte := StrToInt('$' + Copy(S, Lp * 2 - 1, 2));
        Mem.Write(HiByte, 1);
      end;
      // ----------------------------------------------- SHIFT JISɕϊ ---
      Mem.Seek(0, soBeginning);
      ShiftIn := False;
      Result := '';
      repeat
        Mem.Read(HiByte, 1);
        // .............................. GXP[vV[PX̃`FbN ...
        if HiByte = $1B then
        begin
          Mem.Read(HiByte, 1);
          case HiByte of
            $24:  // Shift-in
              begin
                Mem.Read(HiByte, 1);
                if HiByte = $28 then Mem.Seek(1, soCurrent);
                Mem.Read(HiByte, 1);
                ShiftIn := True;
              end;
            $26:  // Shift-in
              begin
                Mem.Seek(4, soCurrent);
                Mem.Read(HiByte, 1);
                ShiftIn := True;
              end;
            $28: // Shift-Out
              begin
                Mem.Seek(1, soCurrent);
                Mem.Read(HiByte, 1);
                ShiftIn := False;
              end;
          end;
        end
        else if HiByte in [$0E, $0F] then
        begin
          Mem.Read(HiByte, 1);
          ShiftIn := (HiByte = $0F);
        end;
        // ............................................. 2oCgR[h ...
        if ShiftIn then
        begin
          Mem.Read(LoByte, 1);
          Dec(HiByte, $21);
          if Odd(HiByte) then Inc(LoByte, $7E) else Inc(LoByte, $1F);
          HiByte := (HiByte and $FE) shr 1;
          if HiByte >= $1F then Inc(HiByte, $C1) else Inc(HiByte, $81);
          if LoByte in [$7F..$9D] then  Inc(LoByte, 1);
          Result := Result + Chr(HiByte) + Chr(LoByte);
        end
        else
          Result := Result + Chr(HiByte);
      until Mem.Position >= Mem.Size;
    finally
      Mem.Free;
    end;
  end
  // ========================================================= Undefined ===
  else
    Result := '';
end;

// ############################### tB[h̏<ClearFields\bh> #####
procedure TExifInfo.ClearFields;
begin
  FApertureValue.RationalText := '';
  FApertureValue.Value := 0;
  FBrightness.RationalText := '';
  FBrightness.Value := 0;
  FColorSpace.Text := '';
  FColorSpace.Value := 0;
  FCompression.Text := '';
  FCompression.Value := 0;
  FCompressBitsPerPixel.RationalText := '';
  FCompressBitsPerPixel.Value := 0;
  FCopyright := '';
  FDateTime := '';
  FDateTimeDigitized := '';
  FDateTimeOriginal := '';
  FExifVersion := '';
  FExposureBias.RationalText := '';
  FExposureBias.Value := 0;
  FExposureProgram.Text := '';
  FExposureProgram.Value := 0;
  FExposureTime.RationalText := '';
  FExposureTime.Value := 0;
  FFlash.Text := '';
  FFlash.Value := 0;
  FFlashPixVersion := '';
  FFNumber.RationalText := '';
  FFNumber.Value := 0;
  FFocalLength.RationalText := '';
  FFocalLength.Value := 0;
  FFocalPlaneResolutionUnit.Text := '';
  FFocalPlaneResolutionUnit.Value := 0;
  FFocalPlaneXResolution.RationalText := '';
  FFocalPlaneXResolution.Value := 0;
  FFocalPlaneYResolution.RationalText := '';
  FFocalPlaneYResolution.Value := 0;
  FImageDescription := '';
  FImageHeight := 0;
  FImageWidth := 0;
  FISOSpeedRating := 0;
  FJpegInterchangeFormat := 0;
  FJpegInterchangeLength := 0;
  FLightSource.Text := '';
  FLightSource.Value := 0;
  FMake := '';
  FMakerNotePosition := 0;
  FMaxAperture.RationalText := '';
  FMaxAperture.Value := 0;
  FMeteringMode.Text := '';
  FMeteringMode.Value := 0;
  FModel := '';
  FOrientation := eoNone;
  FPhotometricInterpretation.Text := '';
  FPhotometricInterpretation.Value := 0;
  FPlanarConfiguration := 0;
  FRelatedSoundFile := '';
  FResolutionUnit.Text := '';
  FResolutionUnit.Value := 0;
  FSamplesPerPixel := 0;
  FSensingMethod.Text := '';
  FSensingMethod.Value := 0;
  FShutterSpeed.RationalText := '';
  FShutterSpeed.Value := 0;
  FSoftware := '';
  FSubjectDistance.RationalText := '';
  FSubjectDistance.Value := 0.0;
  FSubSecTime := '';
  FSubSecTimeDigitized := '';
  FSubSecTimeOriginal := '';
  FThmImageHeight := 0;
  FThmImageWidth := 0;
  FThmOrientation := eoNone;
  FThmResolutionUnit.Text := '';
  FThmResolutionUnit.Value := 0;
  FThmXResolution.RationalText := '';
  FThmXResolution.Value := 0;
  FThmYResolution.RationalText := '';
  FThmYResolution.Value := 0;
  FUserComment.Clear;
  FXResolution.RationalText := '';
  FXResolution.Value := 0;
  FYResolution.RationalText := '';
  FYResolution.Value := 0;
end;

// ######################################## RXgN^<Create\bh> #####
constructor TExifInfo.Create(AOwner: TComponent);
begin
  inherited;
  // ************************************************** tB[h ***
  FUserComment := TStringList.Create;
  ClearFields;
  FImageFileName := '';
  FSearchRange := 32;
end;

// #############################################################################
// #### fXgN^: Destroy\bh
destructor TExifInfo.Destroy;
begin
  FUserComment.Free;
  inherited;
end;

// #############################################################################
// #### vpeB: UserComment 擾
function TExifInfo.GetUserComment: TStrings;
begin
  Result := FUserComment;
end;

// #############################################################################
// ## \bhFExif̃Rs[
procedure TExifInfo.CopyExif(const Dest: string);
const
  ExifHead: array[0..5] of Char = ('E', 'x', 'i', 'f', #0, #0);
var
  BufByte: Byte;
  BufLong: LongWord;
  BufWord: Word;
  EntryCount: Cardinal;   // Gg
  IfdDataSize: Word;      // IFD̃f[^TCY
  Jpg: TJPEGImage;
  L: Integer;
  Mem, Dt: TMemoryStream;
  StartOffset: Integer;        // ItZbg
  Pic: TFileStream;
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // ++ Tu葱FGg̏
  procedure WriteEntry(EntryTag, WFormat: Word; DataCount: LongWord);
  var
    Buffer: Word;
  begin
    Dec(EntryCount);
    Buffer := UnsignedMakeTag(EntryTag, 2);
    Mem.WriteBuffer(Buffer, 2);
    BufWord := UnsignedMakeTag(WFormat, 2);
    Mem.WriteBuffer(BufWord, 2);
    BufLong := UnsignedMakeTag(DataCount, 4);
    Mem.WriteBuffer(BufLong, 4);
  end;
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // ++ Tu֐F̏
  function WriteString(Stream: TStream; const Text: string): Cardinal;
  begin
    Result := Cardinal(Stream.Write(PChar(Text)^, Length(Text) + 1));
  end;
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // ++ Tu葱Ff[^ւ̃ItZbg̏
  procedure WriteDataOffset;
  begin
    BufLong := Cardinal(Mem.Position - StartOffset + Dt.Size) + 12 * EntryCount +
               8;
    BufLong := UnsignedMakeTag(BufLong, 4);
    Mem.WriteBuffer(BufLong, 4);
  end;
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // ++ Tu֐Ff[^̏
  function WriteStringData(const Text: string; ALen: Integer = -1): Cardinal;
  var
    Buffer: Byte;
    Leng: Integer;
    Lp: Integer;
  begin
    Leng := Length(Text) + 1;
    if (Leng <= 4) and (ALen <= 4) then
    begin
      Result := WriteString(Mem, Text);
      if Result < 4 then
      begin
        Buffer := 0;
        for Lp := Result + 1 to 4 do
          Mem.WriteBuffer(Buffer, 1);
        Result := 0;
      end;
    end
    else
    begin
      WriteDataOffset;
      Result := WriteString(Dt, Text);
      if Result < Cardinal(ALen) then
      begin
        Buffer := 0;
        for Lp := Result + 1 to ALen do
          Mem.WriteBuffer(Buffer, 1);
      end;
    end;
  end;
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // ++ Tu葱FUnsigned Short f[^̏
  procedure WriteUShort(Value: Word);
  begin
    BufWord := UnsignedMakeTag(Value, 2);
    Mem.WriteBuffer(BufWord, 2);
    BufWord := 0;
    Mem.WriteBuffer(BufWord, 2);
  end;
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // ++ Tu֐FUnsigned rational f[^̏
  function WriteURational(Value: TExifRational): Cardinal;
  var
    Cp: Integer;
    S1, S2: string;
  begin
    WriteDataOffset;
    Cp := AnsiPos('/', Value.RationalText);
    if Cp > 0 then
    begin
      S1 := Copy(Value.RationalText, 1, Cp - 1);
      S2 := Copy(Value.RationalText, Cp + 1, Length(Value.RationalText) - Cp);
    end
    else
    begin
      S1 := Value.RationalText;
      S2 := '';
    end;
    BufLong := StrToIntDef(S1, 0);
    BufLong := UnsignedMakeTag(BufLong, 4);
    Result := Cardinal(Dt.Write(BufLong, 4));
    BufLong := StrToIntDef(S2, 0);
    BufLong := UnsignedMakeTag(BufLong, 4);
    Result := Result + Cardinal(Dt.Write(BufLong, 4));
  end;
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // ++ Tu葱FUndefined f[^𕶎Ƃď
  procedure WriteUndefText(Stream: TStream; const Text: string; ALen: Integer);
  var
    Buffer: Byte;
    Lp: Integer;
  begin
    if Length(Text) > 0 then Stream.WriteBuffer(PChar(Text)^, Length(Text));
    Buffer := 0;
    for Lp := Length(Text) + 1 to ALen do
      Stream.WriteBuffer(Buffer, 1);
  end;
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // ++ Tu葱FSigned rational f[^̏
  procedure WriteSRational(Value: TExifRational);
  var
    BufSLong: LongInt;
    Cp: Integer;
    S1, S2: string;
  begin
    WriteDataOffset;
    Cp := AnsiPos('/', Value.RationalText);
    if Cp > 0 then
    begin
      S1 := Copy(Value.RationalText, 1, Cp - 1);
      S2 := Copy(Value.RationalText, Cp + 1, Length(Value.RationalText) - Cp);
    end
    else
    begin
      S1 := Value.RationalText;
      S2 := '';
    end;
    BufSLong := StrToIntDef(S1, 0);
    BufSLong := SignedMakeTag(BufSLong, 4);
    Dt.WriteBuffer(BufSLong, 4);
    BufSLong := StrToIntDef(S2, 0);
    BufSLong := SignedMakeTag(BufSLong, 4);
    Dt.WriteBuffer(BufSLong, 4);
  end;
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // ++ Tu葱FUnsignedLong f[^̏
  procedure WriteULong(Value: LongWord);
  begin
    BufLong := UnsignedMakeTag(Value, 4);
    Mem.WriteBuffer(BufLong, 4);
  end;
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
begin
  // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%JPEGt@C̊mF%%
  if not FileExists(Dest) then
    raise EFOpenError.CreateFmt(EMSG_OPEN, [Dest])
  else
  begin
    Pic := TFileStream.Create(Dest, fmOpenRead);
    try
      Pic.Seek(0, soBeginning);
      Pic.ReadBuffer(BufWord, 2);
      if BufWord <> $D8FF then raise EFOpenError.CreateFmt(EMSG_NOTJPEG, [Dest]);
    finally
      Pic.Free;
    end;
  end;
  // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%Exift@C̍쐬%%
  Mem := TMemoryStream.Create;
  try
    // ::::::::::::::::::::::::::::::::::::::::::Exifwb_̍쐬::
    // Offset := 16;
    for L := 0 to 5 do
      Mem.WriteBuffer(ExifHead[L], 1);
    // ::::::::::::::::::::::::::::::::::::::::::TIFFwb_̍쐬::
    // ************************************************i[`**
    StartOffset := Mem.Position;
    if FWriteType = wtII then BufWord := $4949 else BufWord := $4D4D;
    Mem.WriteBuffer(BufWord, 2);
    // **********************************************ItZbg**
    BufWord := $002A;
    BufWord := UnsignedMakeTag(BufWord, 2);
    Mem.WriteBuffer(BufWord, 2);
    BufLong := 8;
    BufLong := UnsignedMakeTag(BufLong, 4);
    Mem.WriteBuffer(BufLong, 4);
    // *****************************************TIFFf[^(IFD)**
    // xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxIFD0xx
    // ==============================fBNgGg==
    BufWord := 6;
    if FImageDescription <> '' then Inc(BufWord);
    if FXResolution.Value <> 0.0 then Inc(BufWord);
    if FYResolution.Value <> 0.0 then Inc(BufWord);
    if FResolutionUnit.Value <> 0 then Inc(BufWord);
    if FCopyright <> '' then Inc(BufWord);
    EntryCount := BufWord;
    BufWord := UnsignedMakeTag(BufWord, 2);
    Mem.WriteBuffer(BufWord, 2);
    // Offset := Integer(EntryCount) * 12 + 2 + Offset;
    // ================================fBNgGg==
    Dt := TMemoryStream.Create;
    try
      // --------------------------------ImageDescription--
      if FImageDescription <> '' then
      begin
        WriteEntry($010E, 2, Cardinal(Length(FImageDescription) + 1));
        WriteStringData(FImageDescription);
      end;
      // --------------------------------------------Make--
      WriteEntry($010F, 2, Cardinal(Length(FMake) + 1));
      WriteStringData(FMake);
      // -------------------------------------------Model--
      WriteEntry($0110, 2, Cardinal(Length(FModel) + 1));
      WriteStringData(FModel);
      // -------------------------------------Orientation--
      WriteEntry($0112, 3, 1);
      WriteUShort(Word(Ord(FOrientation)));
      // -------------------------------------XResolution--
      if FXResolution.Value <> 0.0 then
      begin
        WriteEntry($011A, 5, 1);
        WriteURational(FXResolution);
      end;
      // -------------------------------------YResolution--
      if YResolution.Value <> 0.0 then
      begin
        WriteEntry($011B, 5, 1);
        WriteURational(FYResolution);
      end;
      // ----------------------------------ResolutionUnit--
      if FResolutionUnit.Value <> 0 then
      begin
        WriteEntry($0128, 3, 1);
        WriteUShort(Word(FResolutionUnit.Value));
      end;
      // ----------------------------------------Software--
      WriteEntry($0131, 2, Cardinal(Length(FSoftware) + 1));
      WriteStringData(FSoftware);
      // ----------------------------------------Datetime--
      WriteEntry($0132, 2, 20);
      WriteStringData(FDateTime, 20);
      // ---------------------------------------Copyright--
      if FCopyright <> '' then
      begin
        WriteEntry($8298, 2, Cardinal(Length(FCopyright) + 1));
        WriteStringData(FCopyright);
      end;
      // ----------------------------------ExifIFDPointer--
      WriteEntry($8769, 4, 1);
      BufLong := Cardinal(Mem.Position - StartOffset + Dt.Size) + 8;
      WriteULong(BufLong);
      // ------------------------------IFD1ւ̃ItZbg--
      BufLong := 0;
      Mem.WriteBuffer(BufLong, 4);
      // ----------------------------ǉf[^̏--
      Mem.CopyFrom(Dt, 0);
    finally
      Dt.Free;
    end;
    // xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxExifSubIFDxx
    // ==========================================Gg==
    BufWord := 14;
    if FExposureTime.RationalText <> '' then Inc(BufWord);
    if FFNumber.RationalText <> '' then Inc(BufWord);
    if FExposureProgram.Text <> '' then Inc(BufWord);
    if FCompressBitsPerPixel.RationalText <> '' then Inc(BufWord);
    if FShutterSpeed.RationalText <> '' then Inc(BufWord);
    if FApertureValue.RationalText <> '' then Inc(BufWord);
    if FBrightness.RationalText <> '' then Inc(BufWord);
    if FExposureBias.RationalText <> '' then Inc(BufWord);
    if FMaxAperture.RationalText <> '' then Inc(BufWord);
    if FUserComment.Text <> '' then Inc(BufWord);
    if FSubSecTime <> '' then Inc(BufWord);
    if FSubSecTimeOriginal <> '' then Inc(BufWord);
    if FSubSecTimeDigitized <> '' then Inc(BufWord);
    if FFlashPixVersion <> '' then Inc(BufWord);
    if FFocalPlaneXResolution.RationalText <> '' then Inc(BufWord);
    if FFocalPlaneYResolution.RationalText <> '' then Inc(BufWord);
    if FFocalPlaneResolutionUnit.Text <> '' then Inc(BufWord);
    EntryCount := BufWord;
    BufWord := UnsignedMakeTag(BufWord, 2);
    Mem.WriteBuffer(BufWord, 2);
    // Offset := Offset + Integer(EntryCount) * 12 + 2;
    // ============================================Gg==
    Dt := TMemoryStream.Create;
    try
      Dt.Seek(0, soBeginning);
      // ------------------------------------ExposureTime--
      if FExposureTime.RationalText <> '' then
      begin
        WriteEntry($829A, 5, 1);
        WriteURational(FExposureTime);
      end;
      // -----------------------------------------FNumber--
      if FFNumber.RationalText <> '' then
      begin
        WriteEntry($829D, 5, 1);
        WriteURational(FFNumber);
      end;
      // ---------------------------------ExposureProgram--
      if FExposureProgram.Text <> '' then
      begin
        WriteEntry($8822, 3, 1);
        WriteUShort(Word(FExposureProgram.Value));
      end;
      // ----------------------------------ISOSpeedRating--
      WriteEntry($8827, 3, 1);
      WriteUShort(FISOSpeedRating);
      // -------------------------------------ExifVersion--
      WriteEntry($9000, 7, 4);
      WriteUndefText(Mem, FExifVersion, 4);
      // -------------------------------DateTime Original--
      WriteEntry($9003, 2, 20);
      WriteStringData(FDateTimeOriginal, 20);
      // ------------------------------DateTime Digitized--
      WriteEntry($9004, 2, 20);
      WriteStringData(FDateTimeDigitized, 20);
      // ---------------------------CompressedBitPerPixel--
      if FCompressBitsPerPixel.RationalText <> '' then
      begin
        WriteEntry($9102, 5, 1);
        WriteURational(FCompressBitsPerPixel);
      end;
      // -------------------------------ShutterSpeedValue--
      if FShutterSpeed.RationalText <> '' then
      begin
        WriteEntry($9201, 10, 1);
        WriteSRational(FShutterSpeed);
      end;
      // -----------------------------------ApertureValue--
      if FApertureValue.RationalText <> '' then
      begin
        WriteEntry($9202, 5, 1);
        WriteURational(FApertureValue);
      end;
      // ---------------------------------BrightnessValue--
      if FBrightness.RationalText <> '' then
      begin
        WriteEntry($9203, 10, 1);
        WriteSRational(FBrightness);
      end;
      // -------------------------------ExposureBiasValue--
      if FExposureBias.RationalText <> '' then
      begin
        WriteEntry($9204, 10, 1);
        WriteSRational(FExposureBias);
      end;
      // --------------------------------MaxApertureValue--
      if FMaxAperture.RationalText <> '' then
      begin
        WriteEntry($9205, 5, 1);
        WriteURational(FMaxAperture);
      end;
      // ---------------------------------SubjectDistance--
      WriteEntry($9206, 10, 1);
      WriteSRational(FSubjectDistance);
      // ------------------------------------MeteringMode--
      WriteEntry($9207, 3, 1);
      WriteUShort(Word(FMeteringMode.Value));
      // -------------------------------------LightSource--
      WriteEntry($9208, 3, 1);
      WriteUShort(Word(FLightSource.Value));
      // -------------------------------------------Flash--
      WriteEntry($9209, 3, 1);
      WriteUShort(Word(FFlash.Value));
      // -------------------------------------FocalLength--
      WriteEntry($920A, 5, 1);
      WriteURational(FFocalLength);
      // -------------------------------------UserComment--
      if FUserComment.Text <> '' then
      begin
        WriteEntry($9286, 2, Cardinal(Length(FUserComment.Text) + 1));
        WriteStringData(FUserComment.Text);
      end;
      // --------------------------------------SubsecTime--
      if FSubSecTime <> '' then
      begin
        WriteEntry($9290, 2, Cardinal(Length(FSubSecTime) + 1));
        WriteStringData(FSubSecTime);
      end;
      // ------------------------------SubsecTimeOriginal--
      if FSubSecTimeOriginal <> '' then
      begin
        WriteEntry($9291, 2, Cardinal(Length(FSubSecTimeOriginal) + 1));
        WriteStringData(FSubSecTimeOriginal);
      end;
      // -----------------------------SubsecTimeDigitized--
      if FSubSecTimeDigitized <> '' then
      begin
        WriteEntry($9292, 2, Cardinal(Length(FSubSecTimeDigitized) + 1));
        WriteStringData(FSubSecTimeDigitized);
      end;
      // ---------------------------------FlashPixVersion--
      if FFlashPixVersion <> '' then
      begin
        WriteEntry($A000, 7, 4);
        WriteUndefText(Mem, FFlashPixVersion, 4);
      end;
      // --------------------------------------ColorSpace--
      WriteEntry($A001, 3, 1);
      WriteUShort(Word(FColorSpace.Value));
      // ----------------------------------ExifImageWidth--
      Jpg := TJPEGImage.Create;
      try
        Jpg.LoadFromFile(Dest);
        WriteEntry($A002, 4, 1);
        BufLong := Cardinal(Jpg.Width);
        WriteULong(BufLong);
        // -------------------------------ExifImageHeight--
        WriteEntry($A003, 4, 1);
        BufLong := Cardinal(Jpg.Height);
        WriteULong(BufLong);
      finally
        Jpg.Free;
      end;
      // ---------------------------FocalPlaneXResolution--
      if FFocalPlaneXResolution.RationalText <> '' then
      begin
        WriteEntry($A20E, 5, 1);
        WriteURational(FFocalPlaneXResolution);
      end;
      // ---------------------------FocalPlaneYResolution--
      if FFocalPlaneYResolution.RationalText <> '' then
      begin
        WriteEntry($A20F, 5, 1);
        WriteURational(FFocalPlaneYResolution);
      end;
      // ------------------------FocalPlaneResolutionUnit--
      if FFocalPlaneResolutionUnit.Text <> '' then
      begin
        WriteEntry($A210, 3, 1);
        WriteUShort(Word(FFocalPlaneResolutionUnit.Value));
      end;
      // -----------------------------------SensingMethod--
      WriteEntry($A217, 3, 1);
      WriteUShort(Word(FSensingMethod.Value));
      // ----------------------InteroperabilityIFDPointer--
      WriteEntry($A005, 4, 1);
      BufLong := Cardinal(Mem.Position - StartOffset + Dt.Size) + 8;
      WriteULong(BufLong);
      // -----------------------------------------IFDI--
      BufLong := 0;
      Mem.WriteBuffer(BufLong, 4);
      // --------------------------------------f[^ǉ--
      Mem.CopyFrom(Dt, 0);
    finally
      Dt.Free;
    end;
    // xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxInteroperabilityIFDxx
    // ==========================================Gg==
    BufWord := 2;
    EntryCount := BufWord;
    BufWord := UnsignedMakeTag(BufWord, 2);
    Mem.WriteBuffer(BufWord, 2);
    // Offset := Offset + Integer(EntryCount) * 12 + 2;
    // ============================================Gg==
    Dt := TMemoryStream.Create;
    try
      // ---------------------------InteroperabilityIndex--
      WriteEntry($0001, 2, 4);
      WriteStringData('R98', 4);
      // -------------------------InteroperabilityVersion--
      WriteEntry($0002, 7, 4);
      WriteUndefText(Mem, '0100', 4);
      // -----------------------------------------IFDI--
      BufLong := 0;
      Mem.WriteBuffer(BufLong, 4);
    finally
      Dt.Free;
    end;
    // ****************************************APP1f[^TCY**
    IfdDataSize := Word(Mem.Size) + 2;
    // ::::::::::::::::::::::::::::::::::::::::JPEGC[Wǉ::
    Pic := TFileStream.Create(Dest, fmOpenRead);
    try
      Pic.Seek(2, soBeginning);
      Mem.CopyFrom(Pic, Pic.Size - 2);
    finally
      Pic.Free;
    end;
    // ::::::::::::::::::::::::::::::::::::::::Exif摜Ƃĕۑ::
    Pic := TFileStream.Create(Dest, fmCreate);
    try
      BufWord := $D8FF;
      Pic.WriteBuffer(BufWord, 2);
      BufWord := $E1FF;
      Pic.WriteBuffer(BufWord, 2);
      BufByte := Hi(IfdDataSize);
      Pic.WriteBuffer(BufByte, 1);
      BufByte := Lo(IfdDataSize);
      Pic.WriteBuffer(BufByte, 1);
      Pic.CopyFrom(Mem, 0);
    finally
      Pic.Free;
    end;
  finally
    Mem.Free;
  end;
end;

end.
