{ *************************************************************************** }
{                                                                             }
{ Hotzonu Plug-in Custom Script Unit                                          }
{                                                                             }
{ 2001-04 Hotzonu Project, Monazilla Software Foundation 2001                 }
{                                                                             }
{ *************************************************************************** }

unit ZonuCustom;

interface

uses
  Classes, SysUtils, IniFiles, ZonuClass, ZonuCommon, ZonuHttp, Hzbconst,
  Gikobas;

type
  PParamData  = ^TParamData;
  TParamData  = record
    GikoBas : TGikoBas;
    Subject : string;
    ParseThreadSource: string;
    ParseResSource: string;
  end;

  TCustomScript = class(TBBS)
  private
    { private }
    FParseThreadSource: string;
    FParseResSource: string;
    FThreadListType: TThreadListAccessType;
    FThreadDownloadMode: TThreadDownloadMode;
    FNeedHRefLink: boolean;
    FNeedLinkTarget: boolean;
    FRelativeLink: boolean;
  protected
    { protected }
    function GetThreadListType: TThreadListAccessType; override;
    function GetThreadDownloadMode: TThreadDownloadMode; override;
    function GetNeedHRefLink: boolean; override;
    function GetNeedLinkTarget: boolean; override;
    function GetRelativeLink: boolean; override;
    function GetIsPublished: boolean; override;
    procedure ZonuDataThreadList(Sender: TObject; Downloader: TDownloader;
      Data: string; TotalSize, DownloadSize: integer; var MoreData: boolean); override;
    procedure ZonuEndThreadList(Sender: TObject; Downloader: TDownloader;
      Error: boolean); override;
    procedure ZonuThreadListError(Sender: TObject; Downloader: TDownloader;
      ErrorCode: integer; ErrorMsg, LastResponse: string); override;
    procedure ZonuDataRes(Sender: TObject; Downloader: TDownloader;
      Data: string; TotalSize, DownloadSize: integer; var MoreData: boolean); override;
    procedure ZonuEndRes(Sender: TObject; Downloader: TDownloader;
      Error: boolean); override;
    procedure ZonuResError(Sender: TObject; Downloader: TDownloader;
      ErrorCode: integer; ErrorMsg, LastResponse: string); override;
  public
    { public }
    constructor Create(const ABBSID: string); override;
    procedure ParseSubject(subject: string; out FileName: string;
      out Title: string; out Count: integer); override;
    procedure ParseDat(const ALine: string; const index: integer;
      var ARes: TRes; out ASubject: string); override;
    procedure DownloadThreadList(Downloader: TThreadListDownloader;
      const NoCache, Async, InfSave: boolean); override;
    procedure DownloadThreadItem(Downloader: TResDownloader;
      const NoCache, Async, InfSave, AutoPilot: boolean); override;
    procedure SendRes(Uploader: TResUploader); override;
    procedure SendThread(Uploader: TThreadUploader); override;
    function  GetDisplayUrl(ThreadItem: TThreadItem): string; override;
    function  GetDir(const Url: string): string; override;
    function ConvUrl(const Url: string): string; override;  //bata60

    property  ParseThreadSource: string read  FParseThreadSource  write FParseThreadSource;
    property  ParseResSource: string  read  FParseResSource write FParseResSource;
  end;

implementation

{ TCustomScript }

uses
  MonaUtils;


constructor TCustomScript.Create(const ABBSID: string);
var
  ini : TMemIniFile;
begin
  inherited Create(ABBSID);
  ini             :=  TMemIniFile.Create(ConfigPath + INI_BBS);
  if  (ini.ReadString('OPTION', 'THREADLISTTYPE', 'LUMP')  = 'LUMP') then  begin
    FThreadListType   :=  tlLump;
  end else begin
    FThreadListType   :=  tlOneByOne;
  end;
  if  (ini.ReadString('OPTION', 'THREADDOWNLOADMODE', 'TEXT') = 'TEXT')  then  begin
    FThreadDownloadMode :=  dmText;
  end else begin
    FThreadDownloadMode :=  dmCGI;
  end;
  FNeedHRefLink   :=  ini.ReadBool('OPTION', 'NEEDHREFLINK', False);
  FNeedLinkTarget :=  ini.ReadBool('OPTION', 'NEEDLINKTARGET', False);
  FRelativeLink   :=  ini.ReadBool('OPTION', 'RELATIVELINK', False);
  ini.Free;
end;

//
//  Xbhꗗ̎擾@
//
function TCustomScript.GetThreadListType: TThreadListAccessType;
begin
  Result  :=  FThreadListType;
end;

//
//  Xbh̎擾@
//
function TCustomScript.GetThreadDownloadMode: TThreadDownloadMode;
begin
  Result  :=  FThreadDownloadMode;
end;

//
//  http://ւ̃Nǉ
//
function TCustomScript.GetNeedHRefLink: boolean;
begin
  Result  :=  FNeedHRefLink;
end;

//
//  NւTarget=_blankǉ
//
function TCustomScript.GetNeedLinkTarget: boolean;
begin
  Result  :=  FNeedLinkTarget;
end;

//
//  BBSEBU[hŕ\邩ǂ
//
function TCustomScript.GetIsPublished: boolean;
begin
  Result  :=  True;
end;

//
//  Xbhꗗ_E[h
//
procedure TCustomScript.DownloadThreadList(Downloader: TThreadListDownloader;
  const NoCache, Async, InfSave: boolean);
var
  ParamData : PParamData;
  FZonuHttp : TZonuHttp;
  FGikoBas  : TGikoBas;
begin
  New(ParamData);
  FGikoBas  :=  TGikoBas.Create(nil);
  FGikoBas.Source.LoadFromFile(ParseThreadSource);
  FGikoBas.AddVar('DatName',    '');
  FGikoBas.AddVar('ThreadTitle','');
  FGikoBas.AddVar('ResCount',   '');
  ParamData.GikoBas         :=  FGikoBas;
  Downloader.ParamData      :=  ParamData;

  FZonuHttp                 :=  TZonuHttp.Create(nil);
  FZonuHttp.Downloader      :=  Downloader;
  FZonuHttp.OnZonuData      :=  ZonuDataThreadList;
  FZonuHttp.OnZonuEnd       :=  ZonuEndThreadList;
  FZonuHttp.OnZonuError     :=  ZonuThreadListError;
  FZonuHttp.Url             :=  GetSubjectUrl(Downloader.Board);
  FZonuHttp.NoCache         :=  NoCache;
  FZonuHttp.UserAgent       :=  USER_AGENT;
  FZonuHttp.InfSave         :=  InfSave;
  FZonuHttp.Username        :=  Downloader.Board.BasicAuthorization_UserName; //8
  FZonuHttp.Password        :=  Downloader.Board.BasicAuthorization_Password; //8
  FZonuHttp.GetModified(Downloader.Board.CachePath + 'subject.inf', Async);
end;

//
//  Xbhꗗ擾
//
procedure TCustomScript.ZonuDataThreadList(Sender: TObject;
  Downloader: TDownloader; Data: string; TotalSize,
  DownloadSize: integer; var MoreData: boolean);
var
  dl        : TThreadListDownloader;
  ThreadItem: TThreadItem;
  FGikoBas  : TGikoBas;
begin
  inherited;
  dl  :=  TThreadListDownloader(Downloader);
  FGikoBas  :=  PParamData(dl.ParamData).GikoBas;
  FGikoBas.AddVar('Data',       Data);
  FGikoBas.AddVar('ParseOk',    False);
  FGikoBas.AddVar('ParseOk',    False);
  FGikoBas.AddVar('TempData',   dl.Temp);
  FGikoBas.AddVar('OK',         True);
  FGikoBas.exec;
  if  (FGikoBas.GetVar('ParseOk'))  then  begin
    if (FGikoBas.GetVar('DatName') <> '') then  begin
      ThreadItem                    :=  TThreadItem.Create;
      ThreadItem.BBS                :=  Self;
      ThreadItem.Site               :=  dl.Site;
      ThreadItem.BoardID            :=  dl.Board.BoardID;
      ThreadItem.No                 :=  0;
      ThreadItem.DatFileName        :=  FGikoBas.GetVar('DatName') + '.dat';
      ThreadItem.Title              :=  FGikoBas.GetVar('ThreadTitle');
      ThreadItem.AgeCount           :=  StrToIntDef(FGikoBas.GetVar('ResCount'), 0);
      ThreadItem.RemoteDatFileName  :=  GetReadCgiUrl(dl.Board, FGikoBas.GetVar('DatName') + '.dat');
      ThreadItem.LocalDatFileName   :=  dl.Board.CachePath + ThreadItem.DatFileName;
      dl.FireDataEvt(ThreadItem, DownloadSize, TotalSize, MoreData);
      FGikoBas.AddVar('DatName',    '');
      FGikoBas.AddVar('ThreadTitle','');
      FGikoBas.AddVar('ResCount',   '');
    end;
  end;
  dl.Temp := FGikoBas.GetVar('TempData');
end;

//
//  Xbhꗗ擾I
//
procedure TCustomScript.ZonuEndThreadList(Sender: TObject;
  Downloader: TDownloader; Error: boolean);
var
  ParamData:  PParamData;
begin
  ParamData :=  PParamData(Downloader.ParamData);
  ParamData.GikoBas.Free;
  Dispose(ParamData);
  inherited;
end;

//
//  Xbhꗗ擾G[
//
procedure TCustomScript.ZonuThreadListError(Sender: TObject;
  Downloader: TDownloader; ErrorCode: integer; ErrorMsg, LastResponse: string);
begin
  inherited;

end;

//
//  Xbhꗗ̉
//
procedure TCustomScript.ParseSubject(subject: string; out FileName,
  Title: string; out Count: integer);
var
  FGikoBas: TGikoBas;
begin
  inherited;
  FGikoBas  :=  TGikoBas.Create(nil);
  FGikoBas.Source.LoadFromFile(ParseThreadSource);
  FGikoBas.AddVar('Data',       subject);
  FGikoBas.AddVar('DatName',    '');
  FGikoBas.AddVar('ThreadTitle','');
  FGikoBas.AddVar('ResCount',   '');
  FGikoBas.AddVar('ParseOk',    False);
  FGikoBas.AddVar('ParseOk',    False);
  FGikoBas.AddVar('TempData',   '');
  FGikoBas.AddVar('OK',         True);
  FGikoBas.exec;
  if  (FGikoBas.GetVar('ParseOk'))  then  begin
    FileName  :=  FGikoBas.GetVar('DatName') + '.dat';
    Title     :=  FGikoBas.GetVar('ThreadTitle');
    COunt     :=  StrToIntDef(FGikoBas.GetVar('ResCount'), 0);
  end;
  FGikoBas.Free;
end;

//
//  X̉
//
procedure TCustomScript.ParseDat(const ALine: string; const index: integer;
  var ARes: TRes; out ASubject: string);
begin
  inherited;
  Parse2chDat(ALine, Index, ARes, ASubject);
end;

//
//  \URL
//
function TCustomScript.GetDisplayUrl(ThreadItem: TThreadItem): string;
var
  s : string;
begin
  s :=  GetReadCgiUrl(ThreadItem);
  if  (s = '')  then
    s :=  GetDatUrl(ThreadItem);
  Result  :=  s;
end;

//
//  Xbh̃_E[h
//
procedure TCustomScript.DownloadThreadItem(Downloader: TResDownloader;
  const NoCache, Async, InfSave, AutoPilot: boolean);
var
  ParamData : PParamData;
  FZonuHttp : TZonuHttp;
  FGikoBas  : TGikoBas;
begin
  New(ParamData);
  FGikoBas  :=  TGikoBas.Create(nil);
  FGikoBas.Source.LoadFromFile(ParseResSource);
  FGikoBas.AddVar('ResNo',      '');
  FGikoBas.AddVar('ResMail',    '');
  FGikoBas.AddVar('ResName',    '');
  FGikoBas.AddVar('ResDate',    '');
  FGikoBas.AddVar('ResBody',    '');
  FGikoBas.AddVar('ResHost',    '');
  FGikoBas.AddVar('ResID',      '');
  FGikoBas.AddVar('ThreadTitle','');
  ParamData.GikoBas         :=  FGikoBas;
  Downloader.ParamData      :=  ParamData;

  FZonuHttp                 :=  TZonuHttp.Create(nil);
  FZonuHttp.Downloader      :=  Downloader;
  FZonuHttp.OnZonuData      :=  ZonuDataRes;
  FZonuHttp.OnZonuEnd       :=  ZonuEndRes;
  FZonuHttp.OnZonuError     :=  ZonuResError;
  FZonuHttp.Url             :=  Downloader.ThreadItem.RemoteDatFileName;
  FZonuHttp.NoCache         :=  NoCache;
  FZonuHttp.UserAgent       :=  USER_AGENT;
  FZonuHttp.NeedDataEvent   :=  True;
  FZonuHttp.CacheReady      :=  False;
  FZonuHttp.InfSave         :=  InfSave;
  FZonuHttp.Username        :=  Downloader.Board.BasicAuthorization_UserName; //8
  FZonuHttp.Password        :=  Downloader.Board.BasicAuthorization_Password; //8
  FZonuHttp.CacheReady      :=  Downloader.CacheReady;
  Downloader.HttpObj        :=  FZonuHttp;
  if  (GetThreadDownloadMode = dmCGI)  then  begin
    FZonuHttp.GetCgiDiff(ChangeFileExt(Downloader.ThreadItem.LocalDatFileName, '.inf'), Async);
  end else begin
    FZonuHttp.GetTextDiff(ChangeFileExt(Downloader.ThreadItem.LocalDatFileName, '.inf'), Async);
  end;
end;

//
//  X擾
//
procedure TCustomScript.ZonuDataRes(Sender: TObject; Downloader: TDownloader;
  Data: string; TotalSize, DownloadSize: integer; var MoreData: boolean);
var
  dl  : TResDownloader;
  Res : TRes;
  FGikoBas: TGikoBas;
begin
  dl  :=  TResDownloader(Downloader);
  FGikoBas  :=  PParamData(dl.ParamData).GikoBas;
  FGikoBas.AddVar('Data',       Data);
  FGikoBas.AddVar('ParseOk',    False);
  FGikoBas.AddVar('TempData',   dl.Temp);
  FGikoBas.AddVar('OK',         True);
  FGikoBas.exec;

  if  (FGikoBas.GetVar('ParseOk'))  then  begin
    if  (FGikoBas.GetVar('ThreadTitle') <> '') then  begin
      PParamData(dl.ParamData).Subject  :=  FGikoBas.GetVar('ThreadTitle');
    end;
    if  (FGikoBas.GetVar('ResDate') <> '') then  begin
      //TThreadItem(Adapter.SourceObject).DownloadCount := TThreadItem(Adapter.SourceObject).DownloadCount + 1;
      Res       :=  TRes.Create;
      Res.No    :=  StrToIntDef(FGikoBas.GetVar('ResNo'), 0);
      if  (Res.No = 0)  then  begin
        Res.No  :=  dl.ResList.Count + 1;
      end;
      Res.BBS   :=  Self;
      Res.Site  :=  dl.Site;
      Res.Name  :=  FGikoBas.GetVar('ResName');
      Res.Mail  :=  FGikoBas.GetVar('ResMail');
      Res.Date  :=  FGikoBas.GetVar('ResDate');
      Res.Body  :=  FGikoBas.GetVar('ResBody');
      Res.RemoteHost  :=  FGikoBas.GetVar('ResHost');
      Res.ID          :=  FGikoBas.GetVar('ResID');
      Res.ThreadTitle :=  PParamData(dl.ParamData).Subject;
      dl.FireDataEvt(Res, DownloadSize, TotalSize);
      FGikoBas.AddVar('ResNo',      '');
      FGikoBas.AddVar('ResMail',    '');
      FGikoBas.AddVar('ResName',    '');
      FGikoBas.AddVar('ResDate',    '');
      FGikoBas.AddVar('ResBody',    '');
      FGikoBas.AddVar('ResHost',    '');
      FGikoBas.AddVar('ResID',      '');
      FGikoBas.AddVar('ThreadTitle','');
    end;
  end;
  dl.Temp :=  FGikoBas.GetVar('TempData');
end;

//
//  X擾I
//
procedure TCustomScript.ZonuEndRes(Sender: TObject; Downloader: TDownloader;
  Error: boolean);
var
  ParamData:  PParamData;
begin
  ParamData :=  PParamData(Downloader.ParamData);
  ParamData.GikoBas.Free;
  Dispose(ParamData);
  inherited;
end;

//
//  X擾G[
//
procedure TCustomScript.ZonuResError(Sender: TObject; Downloader: TDownloader;
  ErrorCode: integer; ErrorMsg, LastResponse: string);
begin
  inherited;

end;

//
//  
//
procedure TCustomScript.SendRes(Uploader: TResUploader);
var
  FZonuHttp : TZonuHttp;
begin
  FZonuHttp                 :=  TZonuHttp.Create(nil);
  FZonuHttp.Uploader        :=  Uploader;
  FZonuHttp.OnZonuSendEnd   :=  ZonuSendEnd;
  FZonuHttp.OnZonuSendError :=  ZonuSendError;
  FZonuHttp.Url             :=  GetSendUrl(Uploader.Board, Uploader.ThreadItem);
  FZonuHttp.UserAgent       :=  USER_AGENT;
  FZonuHttp.SendReferer     :=  Uploader.BBS.GetReadCgiUrl(Uploader.ThreadItem);
  FZonuHttp.SendContent     :=  Uploader.SendData.ConvSendContent(GetSendParam);
  FZonuHttp.Username        :=  Uploader.Board.BasicAuthorization_UserName; //8
  FZonuHttp.Password        :=  Uploader.Board.BasicAuthorization_Password; //8
  FZonuHttp.Send;
end;

//
//  VX
//
procedure TCustomScript.SendThread(Uploader: TThreadUploader);
var
  FZonuHttp   : TZonuHttp;
begin
  FZonuHttp                 :=  TZonuHttp.Create(nil);
  FZonuHttp.Uploader        :=  Uploader;
  FZonuHttp.OnZonuSendEnd   :=  ZonuSendEnd;
  FZonuHttp.OnZonuSendError :=  ZonuSendError;
  FZonuHttp.Url             :=  GetNewThreadUrl(Uploader.Board);
  FZonuHttp.UserAgent       :=  USER_AGENT;
  FZonuHttp.SendReferer     :=  Uploader.Board.Url;
  FZonuHttp.SendContent     :=  Uploader.SendData.ConvSendContent(GetNewThreadParam);
  FZonuHttp.SendCookie      :=  Uploader.SendData.ConvSendContent(GetNewThreadCookie);
  FZonuHttp.Username        :=  Uploader.Board.BasicAuthorization_UserName; //8
  FZonuHttp.Password        :=  Uploader.Board.BasicAuthorization_Password; //8
  FZonuHttp.Send;
end;

function TCustomScript.GetRelativeLink: boolean;
begin
  Result  :=  FRelativeLink;
end;

function TCustomScript.GetDir(const Url: string): string;
begin
  //JBBS̏ꍇɗL
  //  http://jbbs.shitaraba.com/computer/351/
  //                            ~~~~~~~~RRԂ
  Result  :=  ChooseString(Url, '/', StrCount1('/', Url) - 2);
end;

function TCustomScript.ConvUrl(const Url: string): string;
var
  s, sx: string;
begin
  //bata60
  //JBBS̏ꍇɗL
  if  (SameText(BBSID, 'ShitarabaJbbs')) then  begin
    //'file:///C:/bbs/read.cgi/computer/351/1040452774/49'
    //[READCGI]
    {
    *=http://jbbs.livedoor.jp/bbs/read.cgi/<dir>/<bbs>/<dat>/
    [PARAM]
    PARAM_BBS=BBS
    PARAM_DAT=KEY
    PARAM_START=START
    PARAM_END=END
    PARAM_RES=
    PARAM_NOFIRST=NOFIRST
    PARAM_TRUE=TRUE
    PARAM_FALSE=FALSE
    }
    //'http://jbbs.livedoor.jp/computer/bbs/rawmode.cgi/computer/351/1011680227/247'
    if  (AnsiPos('&', Url) > 0) then  begin
      Result  :=  Url;
    end else begin
      sx  :=  Copy(Url, AnsiPos('read.cgi', Url), Length(Url));
      if  (sx = Url) then  begin
        sx  :=  Copy(Url, AnsiPos('rawmode.cgi', Url), Length(Url));
      end;
      if  (sx = '') or (sx = Url) then  begin
        s   :=  Url;
      end else begin
        s   :=  'http://jbbs.livedoor.jp/bbs/read.cgi' +
                '?BBS=' + ChooseString(sx, '/', 2) +
                '&KEY=' + ChooseString(sx, '/', 3);
        //read.cgi/computer/351/1040452774/49
        if  (StrCount1('/', s) >3)  then  begin
          //3-5
          if  (AnsiPos('-', ChooseString(sx, '/', 4)) = 0)  then  begin
            s   :=  s + '&START=' + StringReplace(ChooseString(sx, '/', 4), 'n', '', []);
          end else begin

            s   :=  s + '&START=' + ChooseString(ChooseString(sx, '/', 4), '-', 0);
            s   :=  s + '&END='   + ChooseString(ChooseString(sx, '/', 4), '-', 1);

          end;
        end;
        //read.cgi/computer/351/1040452774/49
        if  (StrCount1('/', s) >4)  then  begin
          s   :=  s + '&END=' + StringReplace(ChooseString(sx, '/', 5), 'n', '', []);
        end;
      end;
      Result  :=  s;
    end;
  end else begin
    Result  :=  Url;
  end;
end;

end.
