{ *************************************************************************** }
{                                                                             }
{ Hotzonu HTTP Class Library                                                  }
{                                                                             }
{ 2001-04 Hotzonu Project, Monazilla Software Foundation 2001                 }
{                                                                             }
{ *************************************************************************** }

unit ZonuHttp;

interface

uses
  Classes, Forms, Dialogs, jconvert, SysUtils, HttpProt, IniFiles,
  ZonuClass, ZonuConf, Hzbconst;

type
  TZonuHttpList = class;
  TZonuHttp = class;

  //
  //  :EVENT
  //
  TZonuDataEvt  = procedure (Sender: TObject; Downloader: TDownloader;
    Data: string; TotalSize, DownloadSize: integer;
    var MoreData: boolean) of object;
  TZonuEndEvt   = procedure (Sender: TObject; Downloader: TDownloader;
    Error: boolean) of object;
  TZonuErrorEvt = procedure (Sender: TObject; Downloader: TDownloader;
    ErrorCode: integer; ErrorMsg, LastResponse: string) of object;
  TZonuNotFoundEvt  = procedure (Sender: TObject; Downloader: TDownloader;
    ErrorCode: integer; Recvstr: string) of object;

  TZonuSendEndEvt   = procedure (Sender: TObject; Uploader: TUploader;
    Recvstr: string; Error: boolean) of object;
  TZonuSendErrorEvt = procedure (Sender: TObject; Uploader: TUploader;
    ErrorCode: integer; ErrorMsg, Html: string) of object;

  //
  //  :HttpList
  //
  TZonuHttpList = class(TList)
  private
    function GetZonuHttp(index: integer): TZonuHttp;
    procedure SetZonuHttp(index: integer; const Value: TZonuHttp);
  public
    destructor Destroy; override;
    function  FindHttp(const Url: string; out ZonuHttp: TZonuHttp): boolean;
    function  Add(Value: TZonuHttp): integer;
    procedure Delete(Index: integer);
    procedure Clear; override;
    procedure AbortAll;
    property  Items[index: integer]: TZonuHttp read GetZonuHttp write SetZonuHttp; default;
  end;

  //
  //  :Http
  //
  TZonuHttp = class(THTTPCli)
  private
    FUserAgent: string;
    FZonuDataEvt: TZonuDataEvt;
    FZonuEndEvt: TZonuEndEvt;
    FZonuErrorEvt: TZonuErrorEvt;
    FMoreData : boolean;
    FDownloadSize: integer;
    FStream : TMemoryStream;
    FSize:  integer;
    FDownloader:  TDownloader;
    FUploader:  TUploader;
    FDiffText : boolean;
    FContent  : string;
    ForzContent  : string;//14
    FOverString : TStringList;
    FAborted: boolean;
    FAboned:  boolean;
    FInfSave: boolean;
    FInfFileName: string;
    FNeedDataEvent: boolean;
    FCacheReady: boolean;
    FCacheBuff: string;
    FLogReplace: boolean;
    FZonuNotFoundEvt: TZonuNotFoundEvt;
    FSendReferer: string;
    FSendCookie: string;
    FSendContent: string;
    FResponseStream: TStringStream;
    FRecvCookie: string;
    FZonuSendEndEvt: TZonuSendEndEvt;
    FZonuSendErrorEvt: TZonuSendErrorEvt;
    FAsync: boolean;
    FRequestEnd: boolean; // RequestDone2xȂ悤
    FCookies: TStrings;
    procedure DocData(Sender: TObject; Buffer: Pointer; Len: Integer);
    procedure DocEnd(Sender: TObject);
    procedure RequestDone(Sender : TObject; RqType : THttpRequest; Error : Word);
    procedure RequestDone2(Sender : TObject; RqType : THttpRequest; Error : Word);
    procedure FireDocData(Data: string);
    procedure SendDone(Sender: TObject; RqType: THttpRequest; Error: Word);
    procedure SendEnd(Sender: TObject);
    procedure CookieRcvd(Sender: TObject; const Data: string;
      var Accept: Boolean);
    procedure AmezorDone(Sender: TObject; RqType: THttpRequest;
      Error: Word);
    procedure SetCookies(const Value: TStrings);
    function GetCookies: TStrings;
  protected
    procedure ZonuGet(Async: boolean = True);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure GetContent(const InfFileName: string; Async: boolean = True);
    procedure GetModified(const InfFileName: string; Async: boolean = True);
    procedure GetTextDiff(const InfFileName: string; Async: boolean = True);
    procedure GetCgiDiff(const InfFileName: string; Async: boolean = True);
    procedure Send;
    procedure GetText(const InfFileName: string; Async: boolean = True);  // GETHTML擾
    procedure GetAmezor(const InfFileName: string; Async: boolean = True);

    function GetNotAsync: string; overload;
    procedure GetNotAsync(st: TStream); overload;
    function HeadNotAsync: string;

    property  Downloader: TDownloader   read  FDownloader   write FDownloader;
    property  Uploader: TUploader       read  FUploader     write FUploader;
    property  UserAgent: string         read  FUserAgent    write FUserAgent;
    property  Aborted: boolean          read  FAborted;
    property  Aboned: boolean           read  FAboned;
    property  LogReplace: boolean       read  FLogReplace;  //Diff200
    property  InfSave: boolean          read  FInfSave        write FInfSave;
    property  NeedDataEvent: boolean    read  FNeedDataEvent  write FNeedDataEvent;
    property  Content: string           read  FContent;
    property  CacheReady: boolean       read  FCacheReady     write FCacheReady;
    property  SendContent : string      read  FSendContent  write FSendContent;
    property  SendCookie : string       read  FSendCookie   write FSendCookie;
    property  SendReferer: string       read  FSendReferer  write FSendReferer;
    property  RecvCookie: string        read  FRecvCookie   write FRecvCookie;
    property  RecvCookies: TStrings     read  GetCookies    write SetCookies;
    property  DiffText: boolean         read  FDiffText     write FDiffText;

    property  OnZonuData: TZonuDataEvt  read  FZonuDataEvt  write FZonuDataEvt;
    property  OnZonuEnd: TZonuEndEvt    read  FZonuEndEvt   write FZonuEndEvt;
    property  OnZonuError:TZonuErrorEvt read  FZonuErrorEvt write FZonuErrorEvt;
    property  OnZonuNotFound: TZonuNotFoundEvt  read  FZonuNotFoundEvt  write FZonuNotFoundEvt;
    property  OnZonuSendEnd:TZonuSendEndEvt read  FZonuSendEndEvt write FZonuSendEndEvt;
    property  OnZonuSendError: TZonuSendErrorEvt  read  FZonuSendErrorEvt  write FZonuSendErrorEvt;
  end;

var
  ZonuHttpList : TZonuHttpList;

implementation

uses
  Windows, Zonutils;

const
  MSG_ABORTED = 'Connection aborted on request';
  MSG_TIMEOUT = 'Connection timed out';

////////////////////////////////////////////////////////////////////////////////
{ TZonuHttpList }
////////////////////////////////////////////////////////////////////////////////

//
//  ׂĂ̓]𒆎~
//
procedure TZonuHttpList.AbortAll;
var
  i : integer;
begin
  for i :=  Count - 1 downto  0 do  begin
    Items[i].Abort;
  end;
end;

//
//  ǉ
//
function TZonuHttpList.Add(Value: TZonuHttp): integer;
begin
  Result  :=  inherited Add(Value);
end;

//
//  NA
//
procedure TZonuHttpList.Clear;
var
  i : integer;
begin
  for i :=  Count - 1 downto  0 do  begin
    Delete(i);
  end;
  inherited;
end;

//
//  폜
//
procedure TZonuHttpList.Delete(Index: integer);
begin
  inherited Delete(Index);
end;

//
//  fXgN^
//
destructor TZonuHttpList.Destroy;
begin
  Clear;
  inherited;
end;

//
//  
//
function TZonuHttpList.FindHttp(const Url: string;
  out ZonuHttp: TZonuHttp): boolean;
var
  i : integer;
begin
  Result    :=  False;
  ZonuHttp  :=  nil;
  for i :=  0 to  Count - 1 do  begin
    if  (Items[i].URL = Url)  then  begin
      ZonuHttp  :=  Items[i];
      Result    :=  True;
      Break;
    end;
  end;
end;

//
//  擾
//
function TZonuHttpList.GetZonuHttp(index: integer): TZonuHttp;
begin
  Result  :=  TZonuHttp(inherited Items[index]);
end;

//
//  ݒ
//
procedure TZonuHttpList.SetZonuHttp(index: integer;
  const Value: TZonuHttp);
begin
  inherited Items[index]  :=  Value;
end;

////////////////////////////////////////////////////////////////////////////////
{ TZonuHttp }
////////////////////////////////////////////////////////////////////////////////

constructor TZonuHttp.Create(AOwner: TComponent);
begin
  inherited;
  ZonuHttpList.Add(Self);
  FSize             :=  0;
  FUserAgent        :=  USER_AGENT;
  FDiffText         :=  False;
  FOverString       :=  TStringList.Create;
  FAborted          :=  False;
  FAboned           :=  False;
  FInfSave          :=  True;
  FInfFileName      :=  '';
  FNeedDataEvent    :=  True;
  FStream           :=  TMemoryStream.Create;
  RcvdStream        :=  FStream;
  FCacheReady       :=  True;
  FLogReplace       :=  False;
  FAsync            :=  True;
  FRequestEnd       :=  False;
  FCookies          :=  TStringList.Create;
end;

destructor TZonuHttp.Destroy;
var
  i : integer;
begin
  for i :=  0 to  ZonuHttpList.Count - 1  do  begin
    if  (ZonuHttpList[i] = Self)  then  begin
      ZonuHttpList.Delete(i);
      Break;
    end;
  end;
  if  (FStream <> nil)  then  begin
    FStream.Free;
  end;
  if  (FSendStream <> nil)  then  begin
    FSendStream.Free;
  end;
  if  (FResponseStream <> nil)  then  begin
    FResponseStream.Free;
  end;
  FOverString.Free;
  FCookies.Free;
  inherited;
end;

//
//  If-Modfied-Sincet GET
//
procedure TZonuHttp.GetModified(const InfFileName: string;
  Async: boolean = True);
var
  inf : TMemIniFile;
begin
  if  (NoCache) then  begin
    GetContent(InfFileName, Async);
  end else begin
    FInfFileName  :=  InfFileName;
    if  (FileExists(InfFileName)) then  begin
      inf :=  TMemIniFile.Create(InfFileName);
      Self.ModifiedSince  :=  inf.ReadString('LOG', 'LAST_MODIFIED', '');
      if  (ZonuConfig.HttpMode = hmNoneMatch) or
          (ModifiedSince  = '') then
      begin
        Self.ETag         :=  inf.ReadString('LOG', 'ETAG', '');
      end;
      inf.Free;
    end else begin
      if  (not DirectoryExists(ExtractFilePath(InfFileName))) then  begin
        ForceDirectories(ExtractFilePath(InfFileName));
      end;
    end;
    ZonuGet(ASync);
  end;
end;

//
//  ̂ݎ擾(TEXT)
//
procedure TZonuHttp.GetTextDiff(const InfFileName: string; Async: boolean = True);
var
  inf : TMemIniFile;
  Range : integer;
begin
  FAsync  :=  Async;
  FInfFileName  :=  InfFileName;
  if  (FileExists(InfFileName)) then  begin
    inf :=  TMemIniFile.Create(InfFileName);
    if  (not NoCache) then  begin
      Self.ModifiedSince  :=  inf.ReadString('LOG', 'LAST_MODIFIED', '');
      if  (ZonuConfig.HttpMode = hmNoneMatch) or
          (ModifiedSince  = '') then
      begin
        Self.ETag         :=  inf.ReadString('LOG', 'ETAG', '');
      end;
    end;
    Range               :=  inf.ReadInteger('LOG', 'SIZE', 0);
    if  (Range = 0) then  begin
      Range :=  inf.ReadInteger('LOG', 'Size', 0);
    end;
    if  (Range > 0) then  begin
      Self.ContentRangeBegin    :=  IntToStr(Range - 1);
      FDiffText :=  True;
    end;
    inf.Free;
  end;
  ZonuGet(Async);
end;

//
//  ̂ݎ擾(CGI)
//
procedure TZonuHttp.GetCgiDiff(const InfFileName: string; Async: boolean = True);
var
  FCache  : TStringList;
  //inf : TMemIniFile;
  s   : string;
begin
  FAsync  :=  Async;
  FInfFileName  :=  InfFileName;
  {
  if  (FileExists(FInfFileName)) then  begin
    inf :=  TMemIniFile.Create(FInfFileName);
    Self.ModifiedSince        :=  inf.ReadString('LOG', 'LAST_MODIFIED', '');
    inf.Free;
  end;
  }
  //LbVǂ
  //FDiffText :=  True;
  s         :=  ChangeFileExt(FInfFileName, '.dat');
  if  (FileExists(s)) then  begin
    FCache  :=  TStringList.Create;
    FCache.LoadFromFile(s);
    if  (FCache.Count > 0)  then  begin
      if  (TResDownloader(FDownloader).Board.BBS.SendParam.UsePathInfo) then  begin
        //PATHINFO`̏ꍇ
        s :=  TResDownloader(FDownloader).Board.BBS.SendParam.PathInfoReadCGI;
        s :=  StringReplace(s, '<START>', IntToStr(FCache.Count + 1), [rfIgnoreCase]);
        s :=  StringReplace(s, '<END>', '', [rfIgnoreCase]);
        s :=  StringReplace(s, '<NOFIRST>', TResDownloader(FDownloader).Board.BBS.SendParam.PathInfoParamNofirst, [rfIgnoreCase]);
      end else begin
        //PATHINFO`Ȃꍇ
        s :=  '&' + TResDownloader(FDownloader).Board.BBS.SendParam.ParamStart    + '=' +
                    IntToStr(FCache.Count + 1) +
              '&' + TResDownloader(FDownloader).Board.BBS.SendParam.ParamNoFirst  + '=' +
                    TResDownloader(FDownloader).Board.BBS.SendParam.ParamTrue;
      end;
      Url :=  Url + S;
    end;
    FCache.Free;
  end;
  ZonuGet(Async);
end;

procedure TZonuHttp.GetText(const InfFileName: string;
  Async: boolean = True);
var
  inf : TMemIniFile;
begin
  FAsync  :=  Async;
  FInfFileName  :=  InfFileName;
  if  (FileExists(InfFileName)) then  begin
    inf :=  TMemIniFile.Create(InfFileName);
    Self.ModifiedSince  :=  inf.ReadString('LOG', 'LAST_MODIFIED', '');
    if  (ZonuConfig.HttpMode = hmNoneMatch) or
        (ModifiedSince  = '') then
    begin
      Self.ETag         :=  inf.ReadString('LOG', 'ETAG', '');
    end;
    //FDiffText :=  True;
    inf.Free;
  end;
  ZonuGet(Async);
{
  FCacheFileName  :=  CacheFileName;
  s   :=  ChangeFileExt(CacheFileName, '.inf');
  if  (FileExists(s)) then  begin
    inf :=  TMemIniFile.Create(s);
    Self.ModifiedSince        :=  inf.ReadString('LOG', 'LAST_MODIFIED', '');
    inf.Free;
  end;
  //LbVǂ
  FDiffText :=  True;

  if  (ReadCache) then  begin
    if  (FileExists(FCacheFileName)) then  begin
      FCache.LoadFromFile(FCacheFileName);
    end;
  end;
  ZonuGet(Async);
}
end;

//
//  f[^擾
//
procedure TZonuHttp.ZonuGet(Async: boolean);
begin
  if  (ZonuEnviron.LogStrings <> nil) then  begin
    if  (Async) then  begin
      ZonuEnviron.LogStrings.Add('** GET (ASYNC) **');
    end else begin
      ZonuEnviron.LogStrings.Add('** GET **');
    end;
    ZonuEnviron.LogStrings.Add('GET ' + URL);
    ZonuEnviron.LogStrings.Add('Accept: ' + Self.Accept);
    ZonuEnviron.LogStrings.Add('UserAgent: ' + Self.UserAgent);
    if  (Self.AcceptEncoding <> '') then
      ZonuEnviron.LogStrings.Add('Accept-Encoding: ' + Self.AcceptEncoding);
    if  (Self.ETag <> '')  then
      ZonuEnviron.LogStrings.Add('ETag: ' + Self.ETag);
    if  (Self.ModifiedSince <> '')  then
      ZonuEnviron.LogStrings.Add('If-Modified-Since: ' + Self.ModifiedSince);
    if  (Self.ContentRangeBegin <> '')  then
      ZonuEnviron.LogStrings.Add('Range: ' + Self.ContentRangeBegin);
    ZonuEnviron.LogStrings.Add('');
  end;

  FAsync  :=  Async;
  if  (ZonuConfig.ProxyUse) then  begin
    Proxy         :=  ZonuConfig.Proxy;
    ProxyPort     :=  ZonuConfig.ProxyPort;
    ProxyUsername :=  ZonuConfig.ProxyUserName;
    ProxyPassword :=  ZonuConfig.ProxyPassword;
  end;

  OnDocData     :=  DocData;
  OnDocEnd      :=  DocEnd;
  OnRequestDone :=  RequestDone;

  FMoreData     :=  True;
  Agent         :=  UserAgent;
  FDownloadSize :=  0;
  FContent      :=  '';
  ForzContent   :=  ''; //14
  FCacheBuff    :=  '';

  AutoRetry302  :=  False;

  if  (ZonuEnviron.State = asOnline)  and
      (ZonuEnviron.State = asOnline)  then
  begin
    //IC̏ꍇ
    if  (Async) then  begin
      GetAsync;
    end else begin
      Get;
    end;
  end else begin
    if  (Assigned(FZonuEndEvt)) then  begin
      FZonuEndEvt(Self, FDownloader, False);
    end;
  end;
  if  (not ASync) then  begin
    FreeAndNil(Self);
  end;
end;

//
//  f[^擾
//
procedure TZonuHttp.DocData(Sender: TObject; Buffer: Pointer;
  Len: Integer);
var
  s : string;
  x, p_end : PChar;
  MoreData : boolean;
begin
  if  (FDiffText) and (StatusCode = 200)  then  begin
    FAboned       :=  True;
    FLogReplace   :=  True;
    Abort;
    Exit;
  end else
  if  (StatusCode = 416)  then  begin
    FAboned       :=  True;
    FLogReplace   :=  True;
    Abort;
    Exit;
  end else
  if  (Aboned)  then  begin
    FLogReplace   :=  True;
    Abort;
    Exit;
  end;
  FDownloadSize :=  FDownloadSize + Len;
  if  (not NeedDataEvent) then  begin
    //gzipȂǂœro߃Cxgsvȏꍇ
    if  (ContentLength > 0) then  begin
      FZonuDataEvt(Self, FDownloader, '', ContentLength, FDownloadSize, MoreData);
    end else begin
      FZonuDataEvt(Self, FDownloader, '', 0, 0, MoreData);
    end;
  end else begin
    if  (StatusCode >= 200) and (StatusCode < 300) then  begin
      x := PChar(Buffer);
      p_end := x + Len;
      while x < p_end do begin
        if x^ = #0 then begin
          x^ := '*';
        end;
        Inc(x);
      end;
      if  (CacheReady)  then  begin
        s := FCacheBuff + PChar(Buffer);
        FCacheBuff  :=  '';
        //Lenx  :=  Len;
        FireDocData(s);
      end else begin
        FCacheBuff    :=  FCacheBuff + PChar(Buffer);
      end;
    end;
  end;
end;

procedure TZonuHttp.FireDocData(Data : string);
var
  s, sx, ss : string;
  i, y : integer;
  Lenx : integer;
  MoreData : boolean;

  orz: string; //14
  oi: integer; //14
begin
  Lenx  :=  Length(Data);
  s     :=  Data;
  orz   :=  Data;
  if  (FDiffText) and (Length(S) = 1) and (FContent = '') then  begin
    Exit;
  end;
  if  (FDiffText) and (Length(S) > 0) and (FContent = '') then  begin
    if  (S[1] = #10)  then  begin
      S     :=  Copy(S, 2, Length(S));
      orz   :=  Copy(orz, 2, Length(orz));
      Lenx  :=  Lenx - 1;
    end;
  end;
  FSize :=  FSize + Lenx;
  //14
  {
  ss    :=  JConv(Downloader.CharSet, s);
  sx    :=  FContent  + ss;
  }
  orz   :=  ForzContent + orz;
  ss    :=  JConv(Downloader.CharSet, orz);
  sx    :=  ss;
  
  //FDoc  :=  FDoc      + ss;
  while true  do  begin
    i   :=  AnsiPos(#13#10, sx);
    oi  :=  AnsiPos(#13#10, orz);
    if  (i >  AnsiPos(#10, sx)) or (i = 0) then  begin
      i   :=  AnsiPos(#10, sx);
      oi  :=  AnsiPos(#10, orz);
      y := 1;
    end else begin
      y := 2;
    end;
    if  (i > 0) then  begin
      s     :=  Copy(sx, 1, i - 1);
      sx    :=  Copy(sx, i  + y, Length(sx));
      orz   :=  Copy(orz,oi + y, Length(orz));
      if  (FMoreData) then  begin
        if  (Assigned(FZonuDataEvt))  then  begin
          MoreData := True;
          if  (ContentLength > 0) then  begin
            FZonuDataEvt(Self, FDownloader, s, ContentLength, FDownloadSize, MoreData);
          end else begin
            FZonuDataEvt(Self, FDownloader, s, 0, 0, MoreData);
          end;
          FMoreData     :=  MoreData;
          //
          //  ʂɂ͍ŏA100Xbh\Ȃ
          //  100Xbhȍ~̓Cxgʒm
          //  FOverStringɒ߁A_E[hIɈCɗ
          //
          if  (not FMoreData) then  FOverString.Clear;
        end;
      end else begin
        if  (ContentLength > 0) then  begin
          FZonuDataEvt(Self, FDownloader, '', ContentLength, FDownloadSize, MoreData);
        end else begin
          FZonuDataEvt(Self, FDownloader, '', 0, 0, MoreData);
        end;
        FOverString.Add(s);
      end;
    end else begin
      Break;
    end;
  end;
  FContent      :=  sx;
  ForzContent   :=  orz;
end;

//
//  _E[hI
//
procedure TZonuHttp.DocEnd(Sender: TObject);
var
  ms  : TMemoryStream;
  inf : TMemIniFile;
  s   : string;
  orz : string; //14
  ch  : char;
  i  : integer;
  nSize   : integer;
  MoreData : boolean;
begin
  try

    if  (not FCacheReady) then  begin
      while (not  FCacheReady) do  begin
        Sleep(1000);
        Application.ProcessMessages;
        if  (ZonuEnviron.Disconnected)  then  begin
          raise EZonuError.Create('ؒf܂B');
        end;
      end;
    end;
    if  (FCacheBuff <> '') then  begin
      while true  do  begin
        if  (AnsiPos(#10, FCacheBuff) > 0)  then  begin
          FireDocData(Copy(FCacheBuff, 1, AnsiPos(#10, FCacheBuff)));
          FCacheBuff  :=  Copy(FCacheBuff, AnsiPos(#10, FCacheBuff) + 1, Length(FCacheBuff));
        end else begin
          FireDocData(FCacheBuff);
          Break;
        end;
      end;
    end;
    if  (not FMoreData) then  begin
      //
      //  Xbhꗗɏ܂Ȃf[^
      //  ŏo͂
      //
      for i :=  0 to  FOverString.Count - 1 do  begin
        FZonuDataEvt(Self, FDownloader, FOverString[i], 0, 0, MoreData);
        if  (i mod 32) = 0 then Application.ProcessMessages;
      end;
    end;
    //LbVۑ
    if  (not FAborted) then  begin
      if  (Self.AcceptEncoding = 'gzip') or (Copy(Url, Length(Url) - 2, 3) = '.gz')  then  begin
        ms  :=  TMemoryStream(FStream);
        s   :=  Unzip(ms);
      end else begin
        //̓eۑ
        ms  :=  TMemoryStream(FStream);
        ms.Position :=  0;
        s :=  '';
        while (ms.Position < ms.Size) do  begin
          ms.Read(ch, SizeOf(ch));
          s :=  s + ch;
        end;
      end;
      orz       :=  s;  //14
      //R[hϊĕۑ
      try
        s :=  JConv(FDownloader.CharSet, s);
      except
      end;

      TrimNull(s);  //NULL
      //FResponseText :=  s;
      if  (FDiffText) then  begin
        if  (StatusCode = 416)  then  begin
          FAboned :=  True;
        end else
        {
        if  (StatusCode = 200)  then  begin
          FAboned       :=  True;
          FLogReplace   :=  True;
        end else
        }
        if  (Length(s) > 0) and (Self.ContentRangeBegin <> '')  then  begin
          //ځ[`FbN
          if  (S[1] <> #10) then  begin
            //ځ[񂳂Ă܂
            FAboned :=  True;
            {
            MessageDlg('f[^̎擾Ɏs܂B' + #13#10 +
                       'OĎ擾ĂB',
                       mtError, [mbok], 0);
            }
          end else begin
            if  (Length(S) = 1) then  begin
              //擾f[^
              if  (ZonuEnviron.Cached)  then  begin
                inf :=  TMemIniFile.Create(FInfFileName);
                inf.WriteString('LOG', 'ETAG', ETag);
                //inf.WriteString('LOG', 'DATE', ContentDate);
                inf.UpdateFile;
                inf.Free;
              end;
              Exit;
            end else begin
              S   :=  Copy(S, 2, Length(S));
              orz :=  Copy(orz, 2, Length(orz));
            end;
          end;
        end;
      end;
      FContent      :=  s;
      ForzContent   :=  orz;

      if  (StatusCode >= 200) and (StatusCode < 300) then  begin
        if  (ZonuEnviron.Cached) and (InfSave)  then  begin
          if  (ZonuEnviron.Cached)  then  begin
            if  (not DirectoryExists(ExtractFilePath(FInfFileName))) then  begin
              ForceDirectories(ExtractFilePath(FInfFileName));
            end;
            inf :=  TMemIniFile.Create(FInfFileName);
            inf.WriteString('LOG', 'LAST_MODIFIED', LastModified);
            inf.WriteString('LOG', 'ETAG', ETag);
            //inf.WriteString('LOG', 'DATE', ContentDate);
            if  (FDiffText) then  begin
              //擾
              nSize :=  inf.ReadInteger('LOG', 'SIZE', 0);
              if  (Self.ContentRangeBegin <> '')  then  begin
                inf.WriteInteger('LOG', 'SIZE', nSize + ContentLength - 1);
              end else begin
                inf.WriteInteger('LOG', 'SIZE', nSize + ContentLength);
              end;
            end else begin
              if  (Self.AcceptEncoding <> 'gzip') and (Copy(Url, Length(Url) - 2, 3) <> '.gz')  then
              begin
                inf.WriteInteger('LOG', 'SIZE', ContentLength);
              end;
            end;
            inf.UpdateFile;
            inf.Free;
          end;
        end;
      end;
    end;

  except
  end;
end;

//
//  NGXgI
//
procedure TZonuHttp.RequestDone(Sender: TObject; RqType: THttpRequest;
  Error: Word);
var
  ErrorFlag : boolean;
begin
  if  (FRequestEnd) then  begin
    Exit;
  end else begin
    FRequestEnd :=  True;
  end;
  if  (ZonuEnviron.LogStrings <> nil) then  begin
    ZonuEnviron.LogStrings.Add('** GET-RESP **');
    ZonuEnviron.LogStrings.Add(IntToStr(StatusCode) + ': ' + ReasonPhrase);
    ZonuEnviron.LogStrings.Add('RESP ' + URL);
    ZonuEnviron.LogStrings.Add('Last-Modified: ' + Self.LastModified);
    ZonuEnviron.LogStrings.Add('E-Tag: ' + Self.ETag);
    ZonuEnviron.LogStrings.Add('Date: '  + Self.ContentDate);
    if  (Self.AcceptEncoding <> '') then
      ZonuEnviron.LogStrings.Add('Accept-Encoding: ' + Self.AcceptEncoding);
    if  (Self.AcceptRanges <> '') then
      ZonuEnviron.LogStrings.Add('Accept-Range: ' + Self.AcceptRanges + ' ' +
                          Self.ContentRangeBegin + '-' +
                          Self.ContentRangeEnd);
    ZonuEnviron.LogStrings.Add('Content-Length :' + IntToStr(ContentLength));
    ZonuEnviron.LogStrings.Add('');
  end;

  //if  (not FCacheReady) and (FCacheBuff <> '')  then  begin
  if  (not FCacheReady) then  begin
    while (not  FCacheReady) do  begin
      Sleep(1000);
      Application.ProcessMessages;
      if  (ZonuEnviron.Disconnected)  then  begin
        raise EZonuError.Create('ؒf܂B');
      end;
    end;
  end;
  ErrorFlag :=  False;
  if  (FStatusCode = 416) then  FAboned :=  True;
  if  (FAboned) then  begin
    if  (StatusCode = 404) and (AnsiPos(MSG_ABORTED, ReasonPhrase) = 1)  then
    begin
      ErrorFlag :=  False;
    end else begin
      if  (FLogReplace) then  begin
        if  (Assigned(FZonuErrorEvt)) then  begin
          FZonuErrorEvt(Self, FDownloader, -2, 'X擾ɂƎs܂B[hĂ݂ĂB', '');
        end;
      end else begin
        if  (Assigned(FZonuErrorEvt)) then  begin
          FZonuErrorEvt(Self, FDownloader, -1, 'X폜܂BĎ擾ĂB', '');
        end;
      end;
      ErrorFlag :=  True;
    end;
  end else
  if  (StatusCode <> 0) or (FAboned) then  begin
    //
    //** 304 ** Not Modified
    //
    if  (StatusCode = 304)  then  begin
      if  (not FDiffText) then  begin
        //LoadFromCache;
      end;
    end else
    //
    //** 302 / 404 ** Not Found
    //
    if  (StatusCode = 302) or (StatusCode = 404) or (StatusCode = 403)  then  begin
      ErrorFlag :=  True;
      {
      if  (StatusCode = 404) and
          ((AnsiPos(MSG_ABORTED, ReasonPhrase) = 1)  or
           (AnsiPos(MSG_TIMEOUT, ReasonPhrase) = 1)) then
      begin
        if  (Assigned(FZonuErrorEvt)) then  begin
          FZonuErrorEvt(Self, FDownloader,StatusCode, 'Ȃ񂩒qłB[h邩ċNĂB' + #13#10 + ReasonPhrase);
        end;
      }
      {
      if  (StatusCode = 403)  then  begin

      end else
      }
      if  (StatusCode = 404) and (AnsiPos(MSG_ABORTED, ReasonPhrase) = 1)  then
      begin
        ErrorFlag :=  False;
      end else
      if  (StatusCode = 404) and (AnsiPos(MSG_TIMEOUT, ReasonPhrase) = 1)  then
      begin
        if  (Assigned(FZonuErrorEvt)) then  begin
          FZonuErrorEvt(Self, FDownloader,StatusCode, 'Ȃ񂩒qłB[h邩ċNĂB' + #13#10 + ReasonPhrase, '');
        end;
      end else
      if  (Assigned(FZonuNotFoundEvt)) then  begin
        FZonuNotFoundEvt(Self, FDownloader, StatusCode, FLastResponse {FResponseText});
        if  (FAsync)  then  begin
          FreeAndNil(Self); //ICxg𔭐Ȃ
        end;
        Exit;
      end else begin
        if  (Assigned(FZonuErrorEvt)) then  begin
          //FLastResponse;
          FZonuErrorEvt(Self, FDownloader,StatusCode, ReasonPhrase, FLastResponse);
        end;
      end;
    end else
    if  (StatusCode < 200) or (StatusCode > 299) and (not FAborted)  then
    begin //LZ̏ꍇ̓G[
      {^[R[hʏ (@270)
      if  (擾)  then  begin
        OutputDebugString(PChar(IntToStr(ResponseCode)));
        if  (ResponseCode = 304)  then  begin
        @XVȂ
        end else
        if  (ResponseCode = 206)  then  begin
        @E擪oCgs
        @@擾ς݂̃OɒǋL
        @E擪oCgsȊO
        @@RangeOăNGXgĔs
        end else
        if  (ResponseCode = 200)  then  begin
        @擾ς݂̃ÔĂēւ
        end else
        if  (ResponseCode = 416)  then  begin
        @RangeOăNGXgĔs
        end else begin
        @G[
          uĔsvƂȂĂƂ͒PɃG[ɂ
          폜Ď擾[U[CɂĂOKB
        end;
      end;
      }
      ErrorFlag :=  True;
      if  (Assigned(FZonuErrorEvt)) then  begin
        FZonuErrorEvt(Self, FDownloader,StatusCode, ReasonPhrase, FLastResponse);
      end;
      //if  (Assigned(FZonuErrorEvt)) then  begin
      //  FZonuErrorEvt(Self, StatusCode, ReasonPhrase, FAdapter);
      //end;
      //if  (Assigned(FAdapter.OnHttpError))  then  begin
      //  FAdapter.OnHttpError(FAdapter, StatusCode, ReasonPhrase);
      //end;
    end;
  end;
  if  (Self <> nil) then  begin
    if  (Assigned(FZonuEndEvt)) then  begin
      FZonuEndEvt(Self, FDownloader, ErrorFlag);
    end;
    OnZonuEnd :=  nil;
    if  (FAsync)  then  begin
      FreeAndNil(Self);
    end else begin
      if  (FAboned) and (StatusCode = 404)  then  begin
        FStatusCode  :=  0;
      end;
    end;
  end;
end;

function TZonuHttp.GetNotAsync: string;
begin

  if  (ZonuEnviron.LogStrings <> nil) then  begin
    ZonuEnviron.LogStrings.Add('** GET **');
    ZonuEnviron.LogStrings.Add('GET ' + URL);
    ZonuEnviron.LogStrings.Add('Accept: ' + Self.Accept);
    ZonuEnviron.LogStrings.Add('UserAgent: ' + Self.UserAgent);
    if  (Self.AcceptEncoding <> '') then
      ZonuEnviron.LogStrings.Add('Accept-Encoding: ' + Self.AcceptEncoding);
    if  (Self.ETag <> '')  then
      ZonuEnviron.LogStrings.Add('ETag: ' + Self.ETag);
    if  (Self.ModifiedSince <> '')  then
      ZonuEnviron.LogStrings.Add('If-Modified-Since: ' + Self.ModifiedSince);
    if  (Self.ContentRangeBegin <> '')  then
      ZonuEnviron.LogStrings.Add('Range: ' + Self.ContentRangeBegin);
    ZonuEnviron.LogStrings.Add('');
  end;

  FAsync  :=  False;
  if  (ZonuConfig.ProxyUse) then  begin
    Proxy         :=  ZonuConfig.Proxy;
    ProxyPort     :=  ZonuConfig.ProxyPort;
    ProxyUsername :=  ZonuConfig.ProxyUserName;
    ProxyPassword :=  ZonuConfig.ProxyPassword;
  end;

  Agent         :=  UserAgent;
  RcvdStream    :=  FStream;
  AutoRetry302  :=  False;
  OnDocEnd      :=  DocEnd;
  OnRequestDone :=  RequestDone2;
  if  (ZonuEnviron.State = asOnline)  then  begin
      //*!* (FDownloader.State = asOnline)  then
    Get;
    Result  :=  FContent;//FResponseText;
  end;
end;

procedure TZonuHttp.GetNotAsync(st: TStream);
begin
  if  (ZonuEnviron.LogStrings <> nil) then  begin
    ZonuEnviron.LogStrings.Add('** GET **');
    ZonuEnviron.LogStrings.Add('GET ' + URL);
    ZonuEnviron.LogStrings.Add('Accept: ' + Self.Accept);
    ZonuEnviron.LogStrings.Add('UserAgent: ' + Self.UserAgent);
    if  (Self.AcceptEncoding <> '') then
      ZonuEnviron.LogStrings.Add('Accept-Encoding: ' + Self.AcceptEncoding);
    if  (Self.ETag <> '')  then
      ZonuEnviron.LogStrings.Add('ETag: ' + Self.ETag);
    if  (Self.ModifiedSince <> '')  then
      ZonuEnviron.LogStrings.Add('If-Modified-Since: ' + Self.ModifiedSince);
    if  (Self.ContentRangeBegin <> '')  then
      ZonuEnviron.LogStrings.Add('Range: ' + Self.ContentRangeBegin);
    ZonuEnviron.LogStrings.Add('');
  end;

  FAsync  :=  False;
  if  (ZonuConfig.ProxyUse) then  begin
    Proxy         :=  ZonuConfig.Proxy;
    ProxyPort     :=  ZonuConfig.ProxyPort;
    ProxyUsername :=  ZonuConfig.ProxyUserName;
    ProxyPassword :=  ZonuConfig.ProxyPassword;
  end;

  Agent         :=  UserAgent;
  RcvdStream    :=  st;
  AutoRetry302  :=  False;
  OnDocEnd      :=  DocEnd;
  OnRequestDone :=  RequestDone2;
  if  (ZonuEnviron.State = asOnline)  then  begin
      //*!*(FAdapter.State = asOnline)  then
    Get;
  end;
end;

function TZonuHttp.HeadNotAsync: string;
begin
  FAsync  :=  False;
  if  (ZonuConfig.ProxyUse) then  begin
    Proxy         :=  ZonuConfig.Proxy;
    ProxyPort     :=  ZonuConfig.ProxyPort;
    ProxyUsername :=  ZonuConfig.ProxyUserName;
    ProxyPassword :=  ZonuConfig.ProxyPassword;
  end;

  Agent         :=  UserAgent;
  RcvdStream    :=  FStream;
  OnDocEnd      :=  DocEnd;
  if  (ZonuEnviron.State = asOnline)  then  begin
      //*!*(FAdapter.State = asOnline)  then
    Head;
    Result  :=  FContent; //FResponseText;
  end;
end;


procedure TZonuHttp.Send;
begin
  FCookies.Clear;

  if  (ZonuEnviron.LogStrings <> nil) then  begin
    ZonuEnviron.LogStrings.Add('** POST (ASYNC) **');
    ZonuEnviron.LogStrings.Add('POST ' + URL);
    ZonuEnviron.LogStrings.Add('Accept: ' + Self.Accept);
    ZonuEnviron.LogStrings.Add('UserAgent: ' + Self.UserAgent);
    ZonuEnviron.LogStrings.Add('');
  end;

  if  (ZonuConfig.SendProxyUse) then  begin
    Proxy         :=  ZonuConfig.SendProxy;
    ProxyPort     :=  ZonuConfig.SendProxyPort;
    ProxyUsername :=  ZonuConfig.SendProxyUserName;
    ProxyPassword :=  ZonuConfig.SendProxyPassword;
  end else begin
    Proxy         :=  '';
    ProxyPort     :=  '';
    ProxyUsername :=  '';
    ProxyPassword :=  '';
  end;
  Agent         :=  UserAgent;
  FSendStream       :=  TStringStream.Create(FSendContent);
  FResponseStream   :=  TStringStream.Create('');
  Cookie        :=  FSendCookie;
  Reference     :=  FSendReferer;
  SendStream    :=  FSendStream;
  RcvdStream    :=  FResponseStream;
  OnSendEnd     :=  SendEnd;
  OnRequestDone :=  SendDone;
  OnCookie      :=  CookieRcvd;
  FRecvCookie   :=  '';
  PostAsync;
end;

procedure TZonuHttp.SendEnd(Sender: TObject);
begin
  //
end;


//
//  ݊
//
procedure TZonuHttp.SendDone(Sender: TObject; RqType: THttpRequest;
  Error: Word);
var
  s : string;
begin
  if  (ZonuEnviron.LogStrings <> nil) then  begin
    ZonuEnviron.LogStrings.Add('** POST-RESP **');
    ZonuEnviron.LogStrings.Add(IntToStr(StatusCode) + ': ' + ReasonPhrase);
    ZonuEnviron.LogStrings.Add('RESP ' + URL);
    ZonuEnviron.LogStrings.Add('RESP ' + URL);
    ZonuEnviron.LogStrings.Add('');
  end;

  if  (StatusCode = 304) or (StatusCode = 404) or (StatusCode = 403)  or (StatusCode = 400)  then  begin
    // OK
    if  (Assigned(FZonuSendEndEvt)) then  begin
      FZonuSendEndEvt(Self, FUploader, FContent, False);
    end;
  end else
  if  (StatusCode = 200)  then  begin
    //G[ʒm
    s :=  Jconv(FUploader.CharSet, FResponseStream.DataString);
    if  (s = '') or (FUploader.BBS.SendResultErr = '') then  begin
      if  (Assigned(FZonuSendEndEvt)) then  begin
        FZonuSendEndEvt(Self, FUploader, '', False);
      end;
    end else begin
      if  (AnsiPos(FUploader.BBS.SendResultErr, s) > 0) then  begin
        if  (Assigned(FZonuSendErrorEvt)) then  begin
          FZonuSendErrorEvt(Self, FUploader, StatusCode, ReasonPhrase, FResponseStream.DataString);
        end;
      end else begin
        if  (Assigned(FZonuSendEndEvt)) then  begin
          //FZonuSendEndEvt(Self, FUploader, S, True);
          FZonuSendEndEvt(Self, FUploader, S, False); //11
        end;
      end;
    end;
  end else begin
    //̑̃G[
    FUploader.FireDownloadError(StatusCode, ReasonPhrase, FLastResponse);
  end;
  FreeAndNil(FSendStream);
  FreeAndNil(FResponseStream);
  Free;
end;


procedure TZonuHttp.CookieRcvd(Sender: TObject; const Data: string;
  var Accept: Boolean);
begin
  //Cookie  :=  Data;
  FRecvCookie :=  FRecvCookie + Data;
  Accept  :=  True;

  FCookies.Add(Data);    
end;

procedure TZonuHttp.GetAmezor(const InfFileName: string; Async: boolean = True);
begin
  FAsync  :=  Async;
  FInfFileName  :=  InfFileName;
  if  (ZonuConfig.ProxyUse) then  begin
    Proxy         :=  ZonuConfig.Proxy;
    ProxyPort     :=  ZonuConfig.ProxyPort;
    ProxyUsername :=  ZonuConfig.ProxyUserName;
    ProxyPassword :=  ZonuConfig.ProxyPassword;
  end;
  Agent             :=  UserAgent;
  FSendStream       :=  TStringStream.Create(FSendContent);
  FResponseStream   :=  TStringStream.Create('');
  SendStream        :=  FSendStream;
  RcvdStream        :=  FResponseStream;
  OnRequestDone     :=  AmezorDone;
  if  (FAsync)  then  begin
    PostAsync;
  end else begin
    Post;
  end;
end;

procedure TZonuHttp.AmezorDone(Sender: TObject; RqType: THttpRequest;
  Error: Word);
var
  s : string;
begin
  s :=  FResponseStream.DataString;
  FMoreData :=  True;
  DocData(Self, PChar(s), Length(s));
  DocEnd(Self);
  if  (Assigned(FZonuEndEvt)) then  begin
    FZonuEndEvt(Self, FDownloader, False);
  end;
  FreeAndNil(Self);
end;

procedure TZonuHttp.RequestDone2(Sender: TObject; RqType: THttpRequest;
  Error: Word);
begin
  if  (ZonuEnviron.LogStrings <> nil) then  begin
    ZonuEnviron.LogStrings.Add('** GET-RESP **');
    ZonuEnviron.LogStrings.Add(IntToStr(StatusCode) + ': ' + ReasonPhrase);
    ZonuEnviron.LogStrings.Add('RESP ' + URL);
    ZonuEnviron.LogStrings.Add('Last-Modified: ' + Self.LastModified);
    ZonuEnviron.LogStrings.Add('E-Tag: ' + Self.ETag);
    ZonuEnviron.LogStrings.Add('Date: ' + Self.ContentDate);
    if  (Self.AcceptEncoding <> '') then
      ZonuEnviron.LogStrings.Add('Accept-Encoding: ' + Self.AcceptEncoding);
    if  (Self.AcceptRanges <> '') then
      ZonuEnviron.LogStrings.Add('Accept-Range: ' + Self.AcceptRanges + ' ' +
                          Self.ContentRangeBegin + '-' +
                          Self.ContentRangeEnd);
    ZonuEnviron.LogStrings.Add('Content-Length :' + IntToStr(ContentLength));
    ZonuEnviron.LogStrings.Add('');
  end;
end;

procedure TZonuHttp.GetContent(const InfFileName: string; Async: boolean);
begin
  FInfFileName  :=  InfFileName;
  if  (FileExists(InfFileName)) then  begin
  end else begin
    if  (not DirectoryExists(ExtractFilePath(InfFileName))) then  begin
      ForceDirectories(ExtractFilePath(InfFileName));
    end;
  end;
  ZonuGet(ASync);
end;

function TZonuHttp.GetCookies: TStrings;
begin
  Result  :=  FCookies;
end;

procedure TZonuHttp.SetCookies(const Value: TStrings);
begin
  FCookies.Assign(Value);
end;

initialization
begin
  ZonuHttpList  :=  TZonuHttpList.Create;
end;

finalization
begin
  ZonuHttpList.AbortAll;
  ZonuHttpList.Clear;
  ZonuHttpList.Free;
end;

end.
